from pathlib import Path
from typing import List, Tuple, Union

from pydantic import ConfigDict
from pydantic._internal._config import config_keys
from typing_extensions import TypedDict

MultiPathType = Union[Path, str, List[Union[Path, str]], Tuple[Union[Path, str], ...]]
PathType = Union[Path, str]
DEFAULT_PATH_SENTINEL: PathType = Path("")
STR_SENTINEL: str = ""
SENTINEL = object()
LIST_STR_SENTINEL: list[str] = [""]


class PyProjectConfigDict(TypedDict, total=False):
    """Configuration settings for pyproject.toml."""

    mode_dir_inherit_nested: bool
    mode_dir_inherit_parent: bool
    file_config_inherit_parent: bool
    env_config_inherit_parent: bool

    conf_file: MultiPathType | None
    conf_dir: str | list[str] | None
    conf_ext: str | list[str]
    conf_file_encoding: str | None
    conf_case_sensitive: bool
    conf_ignore_missing: bool
    conf_custom_ext_handler: str | dict[str, str] | None
    conf_include_inherit_parent: list[str]
    conf_exclude_inherit_parent: list[str]

    env_file: str | list[str] | None
    env_prefix: str
    env_prefix_as_mode_dir: bool
    env_prefix_as_nested_mode_dir: bool
    env_prefix_as_source_mode_dir: bool
    env_file_encoding: str | None
    env_case_sensitive: bool
    env_nested_delimiter: str
    env_ignore_missing: bool
    env_include_inherit_parent: list[str]
    env_exclude_inherit_parent: list[str]

    case_sensitive: bool
    ignore_missing: bool
    encoding: str | None
    cli: bool
    secrets_dir: str | None

    handler: str
    handler_inherit_parent: bool
    ordered_settings: list[str]
    ordered_settings_inherit_parent: bool

    include_inherit_parent: list[str]
    exclude_inherit_parent: list[str]

    arfi_debug: bool


class FileConfigDict(TypedDict, total=False):
    """A TypedDict for configuring file settings."""

    conf_file: MultiPathType | None
    conf_dir: MultiPathType | None
    conf_ext: str | list[str]
    conf_file_encoding: str | None
    conf_case_sensitive: bool | None
    conf_ignore_missing: bool | None
    conf_custom_ext_handler: str | dict[str, str] | None
    conf_include_inherit_parent: list[str] | None
    conf_exclude_inherit_parent: list[str] | None


class EnvConfigDict(TypedDict, total=False):
    """A TypedDict for configuring environment settings."""

    env_file: MultiPathType | None
    env_prefix: str | None
    env_prefix_as_mode_dir: bool | None
    env_prefix_as_nested_mode_dir: bool | None
    env_prefix_as_source_mode_dir: bool | None
    env_file_encoding: str | None
    env_case_sensitive: bool | None
    env_nested_delimiter: str | None
    env_ignore_missing: bool | None
    env_include_inherit_parent: list[str] | None
    env_exclude_inherit_parent: list[str] | None


class GlobalConfigDict(TypedDict, total=False):
    """A TypedDict for configuring settings."""

    case_sensitive: bool
    ignore_missing: bool
    encoding: str | None
    cli: bool
    secrets_dir: PathType | None
    include_inherit_parent: list[str] | None
    exclude_inherit_parent: list[str] | None


class SettingsConfigDict(ConfigDict, FileConfigDict, EnvConfigDict, GlobalConfigDict, total=False):
    """A TypedDict for configuring settings."""


config_keys |= set(SettingsConfigDict.__annotations__.keys())


class SettingsParamsDict(TypedDict, total=False):
    """A TypedDict for configuring parameters settings."""

    read_config: bool
    read_config_force: bool
    mode_dir: PathType | None
    mode_dir_inherit_nested: bool
    mode_dir_inherit_parent: bool
    file_config_inherit_parent: bool
    env_config_inherit_parent: bool

    conf_file: MultiPathType | None
    conf_dir: MultiPathType | None
    conf_ext: str | list[str] | None
    conf_file_encoding: str | None
    conf_case_sensitive: bool
    conf_ignore_missing: bool
    conf_custom_ext_handler: str | dict[str, str] | None
    conf_include_inherit_parent: list[str]
    conf_exclude_inherit_parent: list[str]

    env_file: MultiPathType | None
    env_prefix: str
    env_prefix_as_mode_dir: bool
    env_prefix_as_nested_mode_dir: bool
    env_prefix_as_source_mode_dir: bool
    env_file_encoding: str | None
    env_case_sensitive: bool
    env_nested_delimiter: str
    env_ignore_missing: bool
    env_include_inherit_parent: list[str]
    env_exclude_inherit_parent: list[str]

    case_sensitive: bool
    ignore_missing: bool
    encoding: str | None
    cli: bool
    secrets_dir: PathType | None

    handler: str
    handler_inherit_parent: bool
    ordered_settings: list[str]
    ordered_settings_inherit_parent: bool
    include_inherit_parent: list[str]
    exclude_inherit_parent: list[str]



import os
import functools
import inspect
import typing
from pathlib import Path
from typing import Any, Callable

from pydantic._internal._typing_extra import NONE_TYPES, origin_is_union
from pydantic._internal._utils import (
    IMMUTABLE_NON_COLLECTIONS_TYPES,
    is_model_class,
    lenient_issubclass,
)
from pydantic.aliases import AliasPath
from pydantic_core import PydanticUndefined
from typing_extensions import get_args, get_origin

from .errors import ArFiSettingsError

__all__ = [
    "is_settings",
    "is_pydantic",
    "extract_unique_annotations",
    "create_dict_for_path",
    "search_dict_for_path",
    "allow_json_parse_failure",
    "validate_cli_reader",
    "clean_value",
    "is_descriptor",
]


def is_settings(_class: type) -> bool:
    """Check class is subclass of ArFiSettings."""

    from arfi_settings import ArFiSettings

    if lenient_issubclass(_class, ArFiSettings) and _class is not ArFiSettings:
        return True
    return False


def is_pydantic(_class: type) -> bool:
    """Check class is subclass of BaseModel."""
    return is_model_class(_class)


def extract_unique_annotations(args_types: type[Any] | tuple[type[Any], ...] | None) -> set[type[Any]]:
    """Searches for unique annotations in the tuple of types."""

    annotation_typyes = set()
    if not isinstance(args_types, tuple):
        origin_type = get_origin(args_types)
        if origin_type is None:
            annotation_typyes.add(args_types)
        args_types = get_args(args_types)

    for ann in args_types:
        ann_args = get_args(ann)
        if ann_args:
            ext_ann_args = extract_unique_annotations(ann_args)
            annotation_typyes = annotation_typyes.union(ext_ann_args)
        else:
            annotation_typyes.add(ann)
    return annotation_typyes


def create_dict_for_path(
    alias_path: AliasPath | list[str | int],
    value: Any = PydanticUndefined,
) -> dict[str | int, Any]:
    """Creates a dictionary for the path specified by the alias."""

    if isinstance(alias_path, AliasPath):
        alias_path = alias_path.convert_to_aliases()
    last_value = alias_path[-1]
    if value is not PydanticUndefined:
        last_value = {last_value: value}
    for key_alias in reversed(alias_path[:-1]):
        last_value = {key_alias: last_value}
    return last_value


def search_dict_for_path(
    alias_path: AliasPath | list[str | int],
    search_dict: dict,
    case_sensitive: bool = True,
) -> Any:
    """Searches a dictionary for the path specified by the alias.

    Returns:
        The value at the specified path, or `PydanticUndefined` if the path is not found.
    """
    if isinstance(alias_path, AliasPath):
        alias_path = alias_path.convert_to_aliases()

    assert isinstance(alias_path, list), "`alias_path` must be a AliasPath or list[str | int]"

    value = search_dict
    for key in alias_path:
        if isinstance(value, str):
            return PydanticUndefined

        if case_sensitive:
            try:
                value = value[key]
            except (KeyError, IndexError, TypeError):
                return PydanticUndefined
        else:
            if isinstance(key, str):
                search_value = dict()
                if value and isinstance(value, dict):
                    value_by_origin_key = value.get(key, PydanticUndefined)
                    if value_by_origin_key is not PydanticUndefined:
                        value = value_by_origin_key
                        continue
                    key = key.lower()
                    for k, v in value.items():
                        if isinstance(k, str):
                            k = k.lower()
                        if k not in search_value:
                            search_value[k] = v
                else:
                    search_value = value
                try:
                    value = search_value[key]
                except (KeyError, IndexError, TypeError):
                    return PydanticUndefined
            else:
                try:
                    value = value[key]
                except (KeyError, IndexError, TypeError):
                    return PydanticUndefined

    return value


def allow_json_parse_failure(field, field_name: str | None = None) -> bool:
    """Searches non json serializable fields."""

    origin_type = get_origin(field.annotation)
    if not origin_is_union(origin_type):
        if type(field.annotation) in IMMUTABLE_NON_COLLECTIONS_TYPES:
            return True
        if type(field.annotation) in NONE_TYPES:
            return True
        if isinstance(type(field.annotation), type(typing._LiteralGenericAlias)):
            return True
        if field.annotation is Any:
            return True
    else:
        args_types = get_args(field.annotation)
        if any([type(arg) in IMMUTABLE_NON_COLLECTIONS_TYPES for arg in args_types]):
            return True
        if any([type(arg) in NONE_TYPES for arg in args_types]):
            return True
        if any([isinstance(type(arg), type(typing._LiteralGenericAlias)) for arg in args_types]):
            return True
        if Any in args_types:
            return True
    return False


def validate_cli_reader(cli_reader: Callable) -> Callable:
    """Validate and returns cli reader."""

    if not callable(cli_reader):
        raise ArFiSettingsError(f"`cli_reader` must be callable. Got {type(cli_reader)}")

    if inspect.isfunction(cli_reader):
        # Append `self` as firest parameter to function
        oldsig = inspect.signature(cli_reader)
        params = list(oldsig.parameters.values())
        first_sels_params = inspect.Parameter(
            "self",
            inspect.Parameter.POSITIONAL_ONLY,
            default=inspect.Parameter.empty,
        )
        new_params = [first_sels_params]
        new_params.extend(params)
        sig = oldsig.replace(parameters=new_params)

        @functools.wraps(cli_reader)
        def wrapper(*args, **kwargs):
            ba = sig.bind(*args, **kwargs)
            ba.apply_defaults()
            if inspect.getfullargspec(cli_reader).args:
                return cli_reader(*ba.args[1:], **ba.kwargs)
            else:
                return cli_reader(**ba.kwargs)

        wrapper.__signature__ = sig
        return wrapper

    return cli_reader


def clean_value(data: dict[str, Any] | list[Any]) -> dict[str, Any]:
    new_data = dict()
    if isinstance(data, (list, set, tuple)):
        new_iterable = []
        for item in data:
            ni = clean_value(item)
            new_iterable.append(ni)
        return new_iterable

    assert isinstance(data, dict), f"Unexpected type {type(data)}"

    for k, v in data.items():
        if isinstance(k, str) and k.startswith("_"):
            continue
        elif isinstance(v, Path):
            v = v.as_posix()
        elif isinstance(v, dict):
            v = clean_value(v)
        elif isinstance(data, os.PathLike):
            v = str(v)
        new_data[k] = v
    return new_data


def is_descriptor(obj, name: str):
    """Simple check if an attribute is a descriptor."""

    value = inspect.getattr_static(obj, name)
    if getattr(type(value), "__get__", None):
        return True
    return False



import os

import pytest

import arfi_settings
from arfi_settings.init_config import InitSettings


def pytest_configure(config):
    config.addinivalue_line("markers", "todo: need implementation")
    config.addinivalue_line("markers", "current: for debug current tests")
    config.addinivalue_line("markers", "mode_dir")
    config.addinivalue_line("markers", "descriptors")
    config.addinivalue_line("markers", "schemes")
    config.addinivalue_line("markers", "file_config")
    config.addinivalue_line("markers", "env_config")
    config.addinivalue_line("markers", "readers")
    config.addinivalue_line("markers", "handlers")
    config.addinivalue_line("markers", "settings")
    config.addinivalue_line("markers", "field_type")
    config.addinivalue_line("markers", "field_default_value")
    config.addinivalue_line("markers", "field_default_params")
    config.addinivalue_line("markers", "alias")
    config.addinivalue_line("markers", "utils")
    config.addinivalue_line("markers", "case_sensitive")
    config.addinivalue_line("markers", "env_file")
    config.addinivalue_line("markers", "env")
    config.addinivalue_line("markers", "cli")
    config.addinivalue_line("markers", "secret")
    config.addinivalue_line("markers", "pyproject")


@pytest.fixture
def cwd_to_tmp(tmp_path):
    cwd = os.getcwd()
    os.chdir(tmp_path)

    yield tmp_path

    os.chdir(cwd)


@pytest.fixture
def path_base_dir(tmp_path, monkeypatch):
    call_file = tmp_path / "__init__.py"
    call_file.touch(exist_ok=True)

    class PatchedInitSettings(InitSettings):
        @staticmethod
        def _search_called_file():
            return call_file.as_posix(), 0

    monkeypatch.setattr(arfi_settings.init_config, "InitSettings", PatchedInitSettings)
    init_settings = PatchedInitSettings()
    monkeypatch.setattr(arfi_settings.init_config, "init_settings", init_settings)
    monkeypatch.setattr(arfi_settings.main, "init_settings", init_settings)
    yield


@pytest.fixture
def tests_dir():
    cwd = os.getcwd()
    os.chdir("./tests")

    yield tests_dir

    os.chdir(cwd)


@pytest.fixture
def config_dir(cwd_to_tmp):
    config_dir = cwd_to_tmp / "config"
    config_dir.mkdir(exist_ok=True)
    yield config_dir


@pytest.fixture
def secrets_dir(cwd_to_tmp):
    secrets_dir = cwd_to_tmp / "secrets"
    secrets_dir.mkdir(exist_ok=True)
    yield secrets_dir


@pytest.fixture
def cwd_to_tmp_src_project_settings(tmp_path):
    cwd = os.getcwd()
    nested_cwd = tmp_path / "src" / "project" / "settings"
    nested_cwd.mkdir(parents=True, exist_ok=True)
    os.chdir(nested_cwd)

    yield nested_cwd

    os.chdir(cwd)


@pytest.fixture
def cwd_multi_nested(tmp_path):
    cwd = os.getcwd()
    multi_nested_cwd = tmp_path / "src" / "project" / "settings" / "multi" / "nested" / "directory"
    multi_nested_cwd.mkdir(parents=True, exist_ok=True)
    os.chdir(multi_nested_cwd)

    yield multi_nested_cwd

    os.chdir(cwd)


@pytest.fixture
def empty_pyproject_toml_file(cwd_to_tmp):
    pyproject_toml_file = cwd_to_tmp / "pyproject.toml"
    pyproject_toml_file.touch(exist_ok=True)
    yield pyproject_toml_file


@pytest.fixture
def empty_env_file(cwd_to_tmp):
    conf_file = cwd_to_tmp / ".env"
    conf_file.touch(exist_ok=True)
    yield conf_file


@pytest.fixture
def simple_data_env_file(empty_env_file):
    env_file = empty_env_file
    env_file.write_text(
        """
        PATH_CONFIG_FILE = '.env'
        """
    )
    yield env_file


@pytest.fixture
def empty_prod_env_file(cwd_to_tmp):
    conf_file = cwd_to_tmp / "prod.env"
    conf_file.touch(exist_ok=True)
    yield conf_file


@pytest.fixture
def simple_data_prod_env_file(empty_prod_env_file):
    env_file = empty_prod_env_file
    env_file.write_text(
        """
        PATH_CONFIG_FILE = 'prod.env'
        """
    )
    yield env_file


@pytest.fixture
def empty_config_without_extension(cwd_to_tmp):
    conf_file = cwd_to_tmp / "config"
    conf_file.touch(exist_ok=True)
    yield conf_file


@pytest.fixture
def simple_data_config_without_extension(empty_config_without_extension):
    conf_file = empty_config_without_extension
    conf_file.write_text(
        """
        path_config_file = 'config'
        """
    )
    yield conf_file


@pytest.fixture
def empty_config_toml(cwd_to_tmp):
    conf_file = cwd_to_tmp / "config.toml"
    conf_file.touch(exist_ok=True)
    yield conf_file


@pytest.fixture
def simple_data_config_toml(empty_config_toml):
    conf_file = empty_config_toml
    conf_file.write_text(
        """
        path_config_file = 'config.toml'
        """
    )
    yield conf_file


@pytest.fixture
def empty_config_yaml(cwd_to_tmp):
    conf_file = cwd_to_tmp / "config.yaml"
    conf_file.touch(exist_ok=True)
    yield conf_file


@pytest.fixture
def empty_config_yml(cwd_to_tmp):
    conf_file = cwd_to_tmp / "config.yml"
    conf_file.touch(exist_ok=True)
    yield conf_file


@pytest.fixture
def empty_config_json(cwd_to_tmp):
    conf_file = cwd_to_tmp / "config.json"
    conf_file.touch(exist_ok=True)
    yield conf_file


@pytest.fixture
def empty_config_cnf(cwd_to_tmp):
    conf_file = cwd_to_tmp / "config.cnf"
    conf_file.touch(exist_ok=True)
    yield conf_file


@pytest.fixture
def simple_data_config_cnf(empty_config_cnf):
    conf_file = empty_config_cnf
    conf_file.write_text(
        """
        path_config_file = 'config.cnf'
        """
    )
    yield conf_file


@pytest.fixture
def empty_config_config_toml(config_dir):
    conf_file = config_dir / "config.toml"
    conf_file.touch(exist_ok=True)
    yield conf_file


@pytest.fixture
def simple_data_config_config_toml(empty_config_config_toml):
    conf_file = empty_config_config_toml
    conf_file.write_text(
        """
        path_config_file = 'config/config.toml'
        """
    )
    yield conf_file


@pytest.fixture
def empty_config_config_yaml(config_dir):
    conf_file = config_dir / "config.yaml"
    conf_file.touch(exist_ok=True)
    yield conf_file


@pytest.fixture
def simple_data_config_config_yaml(empty_config_config_yaml):
    conf_file = empty_config_config_yaml
    conf_file.write_text(
        """
        path_config_file: 'config/config.yaml'
        """
    )
    yield conf_file


@pytest.fixture
def empty_config_config_yml(config_dir):
    conf_file = config_dir / "config.yml"
    conf_file.touch(exist_ok=True)
    yield conf_file


@pytest.fixture
def simple_data_config_config_yml(empty_config_config_yml):
    conf_file = empty_config_config_yml
    conf_file.write_text(
        """
        path_config_file: 'config/config.yml'
        """
    )
    yield conf_file


@pytest.fixture
def empty_config_config_json(config_dir):
    conf_file = config_dir / "config.json"
    conf_file.touch(exist_ok=True)
    yield conf_file


@pytest.fixture
def simple_data_config_config_json(empty_config_config_json):
    conf_file = empty_config_config_json
    conf_file.write_text(
        """
        {
            "path_config_file": "config/config.json"
        }
        """
    )
    yield conf_file


@pytest.fixture
def empty_secret_path_config_file(secrets_dir):
    conf_file = secrets_dir / "path_config_file"
    conf_file.touch(exist_ok=True)
    yield conf_file


@pytest.fixture
def simple_data_secret_path_config_file(empty_secret_path_config_file):
    conf_file = empty_secret_path_config_file
    conf_file.write_text(
        """
        secrets/path_config_file
        """
    )
    yield conf_file


@pytest.fixture
def alias_data_config_config_toml(empty_config_config_toml):
    conf_file = empty_config_config_toml
    conf_file.write_text(
        """
        alias_config_data = 'alias_config_data'
        ALIAS_CONFIG_DATA = 'ALIAS_CONFIG_DATA'
        AliasConfigData = "AliasConfigData"
        Alias_Config_Data = "Alias_Config_Data"
        config_data = "config_data"
        CONFIG_DATA = "CONFIG_DATA"
        ConfigData = "ConfigData"
        configData = "configData"

        mode = "mode"
        MODE = "MODE"
        Mode = "Mode"

        alias_path = ["alias_path_1","alias_path_2"]
        OTHER_ALIAS_PATH = ["other_alias_path_1","other_alias_path_2"]
        complex_alias_path = {alias_path = ["alias_path_1","alias_path_2"]}
        OTHER_COMPLEX_ALIAS_PATH = {Alias_Path=["other_alias_path_1","other_alias_path_2"],ALIAS_PATH=['OTHER_ALIAS_PATH_1','OTHER_ALIAS_PATH_2']}

        asd_test = 'asd_test'
        """
    )
    yield conf_file


@pytest.fixture
def mode_prod_data_config_config_toml(empty_config_config_toml):
    conf_file = empty_config_config_toml
    conf_file.write_text(
        """
        mode = "prod"
        """
    )
    yield conf_file


@pytest.fixture
def databese_data_env_file(empty_env_file):
    env_file = empty_env_file
    env_file.write_text(
        """
        db_type = 'mysql'
        DB_TYPE = 'postgres'
        db__db_type = 'mysql'
        DB__DB_TYPE = 'postgres'
        DB__DATABASE = 'DB__DATABASE'
        DB__POSTGRES__DATABASE = 'DB__POSTGRES_DATABASE'
        db__MySQL__databese = 'db__MySQL__databese'
        """
    )
    yield env_file


@pytest.fixture
def env_include_exclude_expected():
    expected = (
        "'env_file', 'env_prefix', 'env_prefix_as_mode_dir', "
        "'env_prefix_as_nested_mode_dir', 'env_prefix_as_source_mode_dir', "
        "'env_file_encoding', 'env_case_sensitive', "
        "'env_nested_delimiter', 'env_ignore_missing', "
        "'env_include_inherit_parent' or 'env_exclude_inherit_parent'"
    )
    return expected


@pytest.fixture
def env_include_exclude_msg():
    msg = (
        "Input should be 'env_file', 'env_prefix', 'env_prefix_as_mode_dir', "
        "'env_prefix_as_nested_mode_dir', 'env_prefix_as_source_mode_dir', "
        "'env_file_encoding', 'env_case_sensitive', 'env_nested_delimiter', "
        "'env_ignore_missing', 'env_include_inherit_parent' or "
        "'env_exclude_inherit_parent'"
    )
    return msg


@pytest.fixture
def conf_include_exclude_expected():
    expected = (
        "'conf_file', 'conf_dir', 'conf_ext', 'conf_file_encoding', "
        "'conf_case_sensitive', 'conf_ignore_missing', "
        "'conf_custom_ext_handler', 'conf_include_inherit_parent' or "
        "'conf_exclude_inherit_parent'"
    )
    return expected


@pytest.fixture
def conf_include_exclude_msg():
    msg = (
        "Input should be 'conf_file', 'conf_dir', 'conf_ext', "
        "'conf_file_encoding', 'conf_case_sensitive', 'conf_ignore_missing', "
        "'conf_custom_ext_handler', 'conf_include_inherit_parent' or "
        "'conf_exclude_inherit_parent'"
    )
    return msg


@pytest.fixture
def include_exclude_expected():
    expected = (
        "'conf_file', 'conf_dir', 'conf_ext', 'conf_file_encoding', "
        "'conf_case_sensitive', 'conf_ignore_missing', "
        "'conf_custom_ext_handler', 'conf_include_inherit_parent', "
        "'conf_exclude_inherit_parent', 'env_file', 'env_prefix', "
        "'env_prefix_as_mode_dir', 'env_prefix_as_nested_mode_dir', "
        "'env_prefix_as_source_mode_dir', 'env_file_encoding', "
        "'env_case_sensitive', 'env_nested_delimiter', "
        "'env_ignore_missing', 'env_include_inherit_parent' or "
        "'env_exclude_inherit_parent'"
    )
    return expected


@pytest.fixture
def include_exclude_msg():
    msg = (
        "Input should be 'conf_file', 'conf_dir', 'conf_ext', "
        "'conf_file_encoding', 'conf_case_sensitive', 'conf_ignore_missing', "
        "'conf_custom_ext_handler', 'conf_include_inherit_parent', "
        "'conf_exclude_inherit_parent', 'env_file', 'env_prefix', "
        "'env_prefix_as_mode_dir', 'env_prefix_as_nested_mode_dir', "
        "'env_prefix_as_source_mode_dir', 'env_file_encoding', "
        "'env_case_sensitive', 'env_nested_delimiter', 'env_ignore_missing', "
        "'env_include_inherit_parent' or 'env_exclude_inherit_parent'"
    )
    return msg



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any

from pydantic import StrictBool, validate_call

from .errors import ArFiSettingsError
from .types import (
    DEFAULT_PATH_SENTINEL,
    LIST_STR_SENTINEL,
    SENTINEL,
    PathType,
)

__all__ = [
    "ABCDescriptor",
    "BaseDirDescriptor",
    "ReadConfigDeskriptor",
    "ReadConfigForceDeskriptor",
    "ReadPyProjectTomlDeskriptor",
    "ModeDirDescriptor",
    "ModeDirInheritNestedDescriptor",
    "ModeDirInheritParentDescriptor",
    "FileConfigInheritParentDescriptor",
    "EnvConfigInheritParentDescriptor",
    "HandlerDescriptor",
    "HandlerInheritParentDescriptor",
    "OrderedSettingsDescriptor",
    "OrderedSettingsInheritParentDescriptor",
]


class ABCDescriptor(ABC):
    """Abstract non-data descriptor."""

    name: str = SENTINEL
    default: Any = SENTINEL

    def __init__(self, default: Any = SENTINEL):
        if default is not SENTINEL:
            self.default = self.validate(default)
        self.private_name = f"_{self.name}"
        self.owner = None
        self._inherited_value = SENTINEL

    def __init_subclass__(cls):
        super().__init_subclass__()
        # skip base class
        if cls.__name__.startswith("Base"):
            return
        try:
            assert cls.name is not SENTINEL, f"Missing or empty `name` attribut in {cls.__name__}"
            assert cls.name and isinstance(
                cls.name, str
            ), f"attribut `name` must be not empty string in class {cls.__name__}"
            assert cls.default is not SENTINEL, f"Missing `default` attribut in class {cls.__name__}"
        except AssertionError as e:
            raise ArFiSettingsError(e) from e

    def __set_name__(self, owner, name):
        self.private_name = f"_{name}"
        self.owner = owner

    def __get__(self, instance, owner=None):
        if owner is None:
            return self
        self.owner = owner
        return self.computed_attr(instance, owner)

    @property
    def inherited_value(self) -> Any:
        """Value from parent."""
        if self._inherited_value is SENTINEL:
            try:
                assert self.owner, "Can not define `inherited_value` without owner"
            except AssertionError as e:
                raise ArFiSettingsError(e) from e
            for base in self.owner.__bases__:
                if hasattr(base, self.name):
                    self._inherited_value = getattr(base, self.name)
                    break
        return self._inherited_value

    @staticmethod
    @abstractmethod
    def validate(value: Any) -> Any:
        """Validate init value."""

    def computed_attr(self, instance, owner):
        if instance is not None:
            return getattr(instance, self.private_name, self.default)
        return getattr(owner, self.private_name, self.default)


class BaseStrictBoolDescriptor(ABCDescriptor):
    """Base strict bool descriptor."""

    @staticmethod
    @validate_call
    def validate(value: StrictBool):
        return value


class ReadConfigDeskriptor(BaseStrictBoolDescriptor):
    """Read config descriptor."""

    name: str = "read_config"
    default: bool = True


class ReadConfigForceDeskriptor(ABCDescriptor):
    """Force read config descriptor."""

    name: str = "read_config_force"
    default: bool | None = None

    @staticmethod
    @validate_call
    def validate(value: StrictBool | None):
        return value


class ReadPyProjectTomlDeskriptor(BaseStrictBoolDescriptor):
    """Read pyproject.toml descriptor."""

    name: str = "read_pyproject_toml"
    default: bool = True


class ModeDirInheritNestedDescriptor(BaseStrictBoolDescriptor):
    """Mode directory nested descriptor."""

    name: str = "mode_dir_inherit_nested"
    default: bool = True


class ModeDirInheritParentDescriptor(BaseStrictBoolDescriptor):
    """Mode directory parent descriptor."""

    name: str = "mode_dir_inherit_parent"
    default: bool = True


class FileConfigInheritParentDescriptor(BaseStrictBoolDescriptor):
    """Inherit parent file_config descriptor."""

    name: str = "file_config_inherit_parent"
    default: bool = True


class EnvConfigInheritParentDescriptor(BaseStrictBoolDescriptor):
    """Inherit parent env_config descriptor."""

    name: str = "env_config_inherit_parent"
    default: bool = True


class HandlerInheritParentDescriptor(BaseStrictBoolDescriptor):
    """Inherit parent handler descriptor."""

    name: str = "handler_inherit_parent"
    default: bool = True


class OrderedSettingsInheritParentDescriptor(BaseStrictBoolDescriptor):
    """Inherit parent ordered_settings descriptor."""

    name: str = "ordered_settings_inherit_parent"
    default: bool = True


class OrderedSettingsDescriptor(ABCDescriptor):
    name: str = "ordered_settings"
    default: list[str] = LIST_STR_SENTINEL

    @staticmethod
    @validate_call
    def validate(value: list[str]):
        return value


class HandlerDescriptor(ABCDescriptor):
    """Inherit parent handler descriptor."""

    name: str = "handler"
    default: str = "default_main_handler"

    @staticmethod
    @validate_call
    def validate(value: str):
        return value


class BaseDirDescriptor(ABCDescriptor):
    """Base dir descriptor."""

    name: str = "base_dir"
    default: PathType | None = None

    @staticmethod
    @validate_call
    def validate(value: PathType | None):
        if isinstance(value, str):
            value = Path(value)
        return value


class ModeDirDescriptor(ABCDescriptor):
    """Mode directory descriptor."""

    name: str = "mode_dir"
    default: PathType | None = DEFAULT_PATH_SENTINEL

    @staticmethod
    @validate_call
    def validate(value: PathType | None):
        if isinstance(value, str):
            try:
                assert not value.startswith("."), "`mode_dir` can't startswith `.`"
                assert not value.startswith("/"), "`mode_dir` can't startswith `/`"
                assert not value.endswith("/"), "`mode_dir` can't endswith `/`"
            except AssertionError as e:
                raise ArFiSettingsError(e) from e
        return value

    def computed_attr(self, instance, owner):
        if instance is not None:
            try:
                assert hasattr(
                    instance, "mode_dir_inherit_nested"
                ), f"""missing required attribute `mode_dir_inherit_nested` in class `{instance.__class__.__name__}` for correct operation `mode_dir`"""
            except AssertionError as e:
                raise ArFiSettingsError(e) from e
            mode_dir_inherit_nested = instance.mode_dir_inherit_nested
            mode_dir_attr = getattr(instance, "_mode_dir_attr", None)
            mode_dir = getattr(instance, self.private_name, self.default)
        else:
            try:
                assert hasattr(
                    owner, "mode_dir_inherit_nested"
                ), f"missing required attribute `mode_dir_inherit_nested` in class `{owner.__name__}` for correct operation `mode_dir`"
            except AssertionError as e:
                raise ArFiSettingsError(e) from e
            mode_dir_inherit_nested = owner.mode_dir_inherit_nested
            mode_dir_attr = None
            mode_dir = getattr(owner, self.private_name, self.default)

        mode_dir = (
            mode_dir
            if mode_dir != DEFAULT_PATH_SENTINEL
            else mode_dir_attr
            if mode_dir_attr is not None
            else self.default
        )

        if not mode_dir_inherit_nested:
            return mode_dir

        if self.inherited_value and self.inherited_value not in (SENTINEL, DEFAULT_PATH_SENTINEL):
            if mode_dir and mode_dir != DEFAULT_PATH_SENTINEL:
                mode_dir = f"{self.inherited_value}/{mode_dir}"
            else:
                mode_dir = self.inherited_value

        return mode_dir



import copy
import functools
import itertools
import json
import warnings
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Literal, Union

from pydantic import AliasChoices, AliasPath, BaseModel
from pydantic._internal._typing_extra import origin_is_union
from pydantic._internal._utils import deep_update
from pydantic_core import PydanticUndefined
from typing_extensions import get_args, get_origin

from .constants import ORDERED_SETTINGS
from .errors import ArFiSettingsError
from .readers import ArFiBaseReader, ArFiReader
from .schemes import (
    EnvConfigSchema,
    FileConfigSchema,
    SettingsConfigSchema,
)
from .types import PathType
from .utils import (
    allow_json_parse_failure,
    create_dict_for_path,
    extract_unique_annotations,
    is_pydantic,
    is_settings,
    search_dict_for_path,
)

if TYPE_CHECKING:
    from .main import ArFiSettings

__all__ = (
    "ArFiHandler",
    "ArFiBaseHandler",
)


class ArFiBaseHandler(ABC):
    """Base handler class."""

    settings_class: "ArFiSettings" = None
    """The passed class for which the settings are being processed."""
    init_kwargs: dict[str, Any]
    """Parameters passed at initialisation or during previous processing."""
    handler: str = "default_main_handler"
    """Handler Method.
    Can be overridden for any instance of the passed settings reading class.
    """
    ordered_settings: list[str] = ORDERED_SETTINGS
    """In what order to read the settings.
    For any instance, you can swap / add new ones / remove existing ones,
    i.e. override as you like, regardless of the value set in the class
    """
    config: SettingsConfigSchema
    """The main settings parameters corresponding to the current class."""
    reader_class: ArFiBaseReader = ArFiReader
    """Settings reader class.
    Can be overridden in a handler that is defined in the passed settings class.
    """
    fields_names: set[str]
    """The names of all fields that are in the passed settings class."""
    fields_is_settings: list[str]
    """Field names that are correctly readable settings.
    There are limitations!!!
    ```
    class SQLite(ArFiSettings):
        pass

    class Postgres(ArFiSettings):
        pass

    class MySQL(BaseModel):
        pass

    # correctly processed fields - are readable settings
    class AppConfig(ArFiSettings):
        db: Postgres

    # correctly processed fields - are readable settings
    class AppConfig(ArFiSettings):
        db: Postgres | SQLite

    # Settings for the `db` specified in the `Postgres` class will not be read
    class AppConfig(ArFiSettings):
        db: Postgres | MySQL  # MySQL inherits from BaseModel, not ArFiSettings

    # Settings for the `db` specified in the `Postgres` class will not be read
    class AppConfig(ArFiSettings):
        db: Postgres | str | None

    # Settings for `db` set in `Postgres` or `SQLite` class will not be read
    class AppConfig(ArFiSettings):
        db: Postgres | SQLite | None

    # Settings for the `db` specified in the `Postgres` class will not be read
    class AppConfig(ArFiSettings):
        db: list[Postgres]
    ```
    """
    field_aliases: dict[str, list[str]]
    """Register-dependent aliases for specific fields.
    The key is the name of the field.
    Sorting is set by Pydantic
    """
    field_aliases_not_case_sensitive: dict[str, list[str]]
    """Registro NOT dependent aliases for specific fields.
    The key is the name of the field.
    Sorting is set by Pydantic,
    but alias.lower() is added after the case-dependent alias.
    """
    alias_fields: dict[str, list[str]]
    """List of fields for case-dependent aliases.
    Key - case-dependent alias
    """
    alias_fields_not_case_sensitive: dict[str, list[str]]
    """List of fields for case-independent aliases.
    Key - alias, case-independent
    """
    fields_alias_path: dict[str, dict[str, int | list[str | int]]]
    """Registry-dependent AliasPath
    Key - alias, case-sensitive
    """
    fields_alias_path_not_case_sensitive: dict[str, dict[str, int | list[str | int | list[str | int]]]]
    """Registro NOT dependent AliasPath
    Key - alias, case NOT dependent
    """
    fields_is_pydantic: list[str]
    """
    Names of fields inherited from the pydantic.BaseModel.
    """
    pydantic_single_fields: dict[str, list[str]]
    """
    Dict of fields keys info.
    Key - field name
    """
    fields_discriminator: dict[str, dict[str, Any]]
    """
    Fields that have a discriminator as a string.
    Key - field name
    """
    pydantic_aliases_info: dict[str, list[dict[str, Any]]]
    """
    Information about nested fields, if the field itself is inherited from pydantic.BaseModel.
    Key - field name
    """
    allowed_json_parse_failure_fields: set[str]
    """
    Fields that can be ignored when parsing JSON.
    """
    data: dict[str, Any]
    """
    Data returned by the handler.
    """

    def __init__(
        self,
        settings_class: "ArFiSettings",
        init_kwargs: dict[str, Any],
        handler: str = "",
    ) -> None:
        super().__init__()
        self.settings_class = settings_class
        self.reader_class.BASE_DIR = settings_class.BASE_DIR
        self.reader_class.ROOT_DIR = settings_class.root_dir
        self.init_kwargs = init_kwargs
        self.config = self.settings_class.settings_config
        if handler:
            self.handler = self._validate_main_handler(handler)
        self.config.conf_ext = self._validate_conf_ext(self.config.conf_ext)
        self.ordered_settings = self._validate_ordered_settings_handlers(
            self.settings_class.ordered_settings,
            self.settings_class,
        )
        self.conf_custom_ext_handler = self._validate_conf_custom_ext_handler(
            self.config.conf_custom_ext_handler,
        )
        self.data = {}
        self.fields_is_settings = []
        self.fields_is_pydantic = []
        self.pydantic_single_fields = dict()
        self.fields_names = set()
        self.field_aliases = dict()
        self.field_aliases_not_case_sensitive = dict()
        self.alias_fields = dict()
        self.alias_fields_not_case_sensitive = dict()
        self.fields_alias_path = dict()
        self.fields_alias_path_not_case_sensitive = dict()
        self.fields_defaults = dict()
        self.fields_discriminator = dict()
        self.pydantic_aliases_info = dict()
        self.allowed_json_parse_failure_fields = set()
        self._extract_fields_info()
        self._extract_alias_info()
        self._prepare_init_kwargs()

    def _prepare_init_kwargs(self) -> None:
        """Separate init and handler kwargs."""

        from arfi_settings import ArFiSettings

        _handler_value: bool = self.init_kwargs.get("_handler_value", False)
        _init_value: dict = self.init_kwargs.get("_init_value", {})

        for field_name, field_value in self.init_kwargs.items():
            if field_name in self.fields_is_settings:
                if issubclass(type(field_value), ArFiSettings):
                    instance_id = id(field_value)
                    field_value = field_value.model_dump()
                    # REMOVE computed fields
                    # field_value.pop("conf_path")
                    # field_value.pop("env_path")
                    field_value["_instance_id"] = instance_id
                    self.init_kwargs[field_name] = field_value
        self.init_kwargs = self._convert_data_to_field_names(self.init_kwargs)
        if not _handler_value:
            for field_name, field_value in self.init_kwargs.items():
                if field_name in self.fields_is_settings and field_value:
                    self.init_kwargs[field_name]["_init_value"] = field_value
        else:
            self.data = copy.deepcopy(self.init_kwargs)
            self.init_kwargs = dict()
            if _init_value:
                _init_value = self._convert_data_to_field_names(_init_value)
                for field_name, field_value in _init_value.items():
                    if field_name in self.fields_is_settings and field_value:
                        _init_value[field_name]["_init_value"] = field_value
                self.init_kwargs = _init_value

    def __init_subclass__(cls):
        super().__init_subclass__()
        if cls.handler != "default_main_handler":
            cls.handler = cls._validate_main_handler(cls.handler)

    def _alias_decorator(self, func):
        """Find value by alias and return it."""

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            valid_data = dict()
            fields_set = set()
            data = func(*args, **kwargs)
            case_sensitive = data.get("__case_sensitive", self.config.case_sensitive)
            lower_data = dict()
            if case_sensitive:
                dict_fields_alias = self.field_aliases
                dict_fields_alias_path = self.fields_alias_path
            else:
                dict_fields_alias = self.field_aliases_not_case_sensitive
                dict_fields_alias_path = self.fields_alias_path_not_case_sensitive
                for key, value in data.items():
                    if isinstance(key, str):
                        key = key.lower()
                    if key not in lower_data:
                        lower_data[key] = value
            for field_name, aliases in dict_fields_alias.items():
                fields_alias_path = dict_fields_alias_path.get(field_name)
                for alias in aliases:
                    if field_name in fields_set:
                        break
                    value = data.get(alias, PydanticUndefined)
                    if not case_sensitive and value is PydanticUndefined:
                        value = lower_data.get(alias, PydanticUndefined)

                    if value is not PydanticUndefined:
                        if fields_alias_path:
                            if alias_path := fields_alias_path.get(alias):
                                if isinstance(value, str):
                                    value = json.loads(value)
                                value_dict = {alias: value}
                                field_value = search_dict_for_path(alias_path, value_dict, case_sensitive)
                                valid_data[field_name] = field_value
                                fields_set.add(field_name)
                            else:
                                valid_data[field_name] = value
                                fields_set.add(field_name)
                        else:
                            valid_data[field_name] = value
                            fields_set.add(field_name)
            return valid_data

        return wrapper

    def __getattribute__(self, name):
        value = object.__getattribute__(self, name)
        if callable(value) and not name.startswith("_"):
            alias_decorator = object.__getattribute__(self, "_alias_decorator")
            if name.endswith("ext_handler"):
                value = alias_decorator(value)
            elif name == "cli_ordered_settings_handler":
                value = alias_decorator(value)
            elif name == "secrets_ordered_settings_handler":
                value = alias_decorator(value)
        return value

    def _convert_data_to_field_names(self, data: dict[str, Any]) -> dict[str, Any]:
        """Renames keys in data from alias to field names."""

        valid_data = dict()
        fields_set = set()
        add_to_fieldset = True
        is_handler_value = data.get("_handler_value", False)
        fields_alias_path = self.fields_alias_path
        if not self.config.case_sensitive:
            fields_alias_path = self.fields_alias_path_not_case_sensitive

        for name, value in data.items():
            if is_handler_value and name not in fields_set:
                # NOTE: Only for AliasPath
                if name in self.fields_names and name in fields_alias_path:
                    valid_data[name] = value

            alias_fields = self.alias_fields.get(name)
            if not alias_fields and not self.config.case_sensitive:
                alias_fields = self.alias_fields_not_case_sensitive.get(name)
                if not alias_fields:
                    add_to_fieldset = False
                    alias_fields = self.alias_fields_not_case_sensitive.get(name.lower())

            if alias_fields:
                for field_name in alias_fields:
                    field_aliases = self.field_aliases.get(field_name)
                    first_alias = field_aliases[0]
                    if name != first_alias and field_name in fields_set:
                        continue
                    if fields_aliases_path := self.fields_alias_path.get(field_name):
                        if alias_path := fields_aliases_path.get(name):
                            field_value = search_dict_for_path(alias_path, {name: value})
                            if field_value is not PydanticUndefined:
                                valid_data[field_name] = field_value
                                if add_to_fieldset:
                                    fields_set.add(field_name)
                        else:
                            valid_data[field_name] = value
                            if add_to_fieldset:
                                fields_set.add(field_name)
                    else:
                        valid_data[field_name] = value
                        if add_to_fieldset:
                            fields_set.add(field_name)

        return valid_data

    def _convert_data_to_valid_aliases(self, data: dict[str, Any]) -> dict[str, Any]:
        """Renames keys in data from field names to valid aliases."""

        valid_data = dict()
        fields_set = set()
        for field_name, aliases in self.field_aliases.items():
            value = data.get(field_name, PydanticUndefined)
            if value is not PydanticUndefined:
                fields_alias_path = self.fields_alias_path.get(field_name)
                for alias in aliases:
                    if field_name in fields_set:
                        break

                    if fields_alias_path:
                        if alias_path := fields_alias_path.get(alias):
                            alias_dict = create_dict_for_path(alias_path, value)
                            if valid_data.get(alias):
                                valid_data[alias].update(alias_dict[alias])
                            else:
                                valid_data.update(alias_dict)
                            fields_set.add(field_name)
                        else:
                            alias_dict = {alias: value}
                            valid_data.update(alias_dict)
                            fields_set.add(field_name)
                    else:
                        alias_dict = {alias: value}
                        valid_data.update(alias_dict)
                        fields_set.add(field_name)
        return valid_data

    def __call__(self) -> dict[str, Any]:
        """Generate settings data."""

        handler = self._get_main_handler(self.handler)
        self.data = handler()

        _handler_tree = copy.deepcopy(self.settings_class._handler_tree)
        _handler_parent_mode_dir = copy.deepcopy(self.settings_class._handler_parent_mode_dir)
        _handler_ordered_settings = copy.deepcopy(self.ordered_settings)
        _handler_parent_file_config = FileConfigSchema(**self.settings_class.computed_file_config)
        _handler_parent_env_config = EnvConfigSchema(**self.settings_class.computed_env_config)

        if self.settings_class.mode_dir is not None:
            _handler_parent_mode_dir.append(self.settings_class.mode_dir)
        read_config_data = {
            "_read_config": True,
            "_handler_value": True,
            "_handler_read_pyproject_toml_force": False,
            "_handler_search_base_dir": False,
            "_handler_parent_mode_dir": _handler_parent_mode_dir,
            "_handler_ordered_settings": _handler_ordered_settings,
            "_handler_parent_file_config": _handler_parent_file_config,
            "_handler_parent_env_config": _handler_parent_env_config,
            "_handler_main_handler": self.settings_class._handler,
        }
        for field in self.fields_is_settings:
            if not self.data.get(field):
                self.data[field] = {}
            self.data[field].update(read_config_data)
            self.data[field]["_handler_mode_dir_attr"] = field
            if self.config.env_case_sensitive:
                field_aliases = self.field_aliases[field]
            else:
                field_aliases = self.field_aliases_not_case_sensitive[field]
            product_tree = copy.deepcopy(_handler_tree)
            product_tree.append(field_aliases)
            self.data[field]["_handler_tree"] = product_tree

            if field in self.fields_defaults:
                instance_id = self.fields_defaults[field]
                self.data[field]["_instance_id"] = instance_id

            # Resolve default discriminator
            if discriminator := self.fields_discriminator.get(field):
                discriminator_key, discriminator_value = list(discriminator.items())[0]
                if not self.config.case_sensitive:
                    if not self.data[field].get(discriminator_key):
                        exist_discriminator_value = PydanticUndefined
                        for key, value in self.data[field].items():
                            if key.lower() == discriminator_key.lower():
                                exist_discriminator_value = value
                                self.data[field].pop(key)
                                break
                        if exist_discriminator_value is not PydanticUndefined:
                            self.data[field][discriminator_key] = exist_discriminator_value
                        elif discriminator_value is not PydanticUndefined:
                            self.data[field][discriminator_key] = discriminator_value
                else:
                    if not self.data[field].get(discriminator_key) and discriminator_value is not PydanticUndefined:
                        self.data[field][discriminator_key] = discriminator_value
        self.data = self._convert_data_to_valid_aliases(self.data)

        return self.data

    def _append_field_to_alias(self, alias: str, field: str) -> None:
        if not self.alias_fields.get(alias):
            self.alias_fields[alias] = []
        if field not in self.alias_fields[alias]:
            self.alias_fields[alias].append(field)

    def _append_alias_to_field(self, field: str, alias: str) -> None:
        if not self.field_aliases.get(field):
            self.field_aliases[field] = []
        if alias not in self.field_aliases[field]:
            self.field_aliases[field].append(alias)
        self._append_field_to_alias(alias, field)
        self._append_alias_not_case_sensitive_to_field(field, alias)

    def _append_field_to_alias_not_case_sensitive(self, alias: str, field: str) -> None:
        if not self.alias_fields_not_case_sensitive.get(alias):
            self.alias_fields_not_case_sensitive[alias] = []
        if field not in self.alias_fields_not_case_sensitive[alias]:
            self.alias_fields_not_case_sensitive[alias].append(field)

    def _append_alias_not_case_sensitive_to_field(self, field: str, alias: str) -> None:
        if not self.field_aliases_not_case_sensitive.get(field):
            self.field_aliases_not_case_sensitive[field] = []
        for als in (alias, alias.lower()):
            if als not in self.field_aliases_not_case_sensitive[field]:
                self.field_aliases_not_case_sensitive[field].append(als)
            self._append_field_to_alias_not_case_sensitive(als, field)

    def _append_fields_alias_path(self, field: str, alias_path: AliasPath | list[str | int]) -> None:
        if isinstance(alias_path, AliasPath):
            alias_path = alias_path.convert_to_aliases()
        first_arg = alias_path[0]
        self._append_alias_to_field(field, first_arg)
        # case_sensitive=True
        if self.fields_alias_path.get(field) is None:
            self.fields_alias_path[field] = dict()
        self.fields_alias_path[field][first_arg] = alias_path

        # case_sensitive=False
        if self.fields_alias_path_not_case_sensitive.get(field) is None:
            self.fields_alias_path_not_case_sensitive[field] = dict()
        self.fields_alias_path_not_case_sensitive[field][first_arg] = alias_path
        self.fields_alias_path_not_case_sensitive[field][first_arg.lower()] = alias_path

    def _extract_alias_info(self) -> None:
        """Extract information about aliases from settings_class."""

        for name, field in self.settings_class.model_fields.items():
            if field.validation_alias is None:
                self._append_alias_to_field(name, name)
                continue

            if isinstance(field.validation_alias, str):
                self._append_alias_to_field(name, field.validation_alias)
                continue

            validation_aliases: list[str | AliasPath] = (
                field.validation_alias.choices
                if isinstance(field.validation_alias, AliasChoices)
                else [field.validation_alias]
            )
            for alias in validation_aliases:
                if isinstance(alias, str):
                    self._append_alias_to_field(name, alias)
                    continue

                assert isinstance(alias, AliasPath)
                self._append_fields_alias_path(name, alias)

    def _extract_fields_info(self) -> None:
        """Extract fields info."""

        from arfi_settings import ArFiSettings

        for field_name, field in self.settings_class.model_fields.items():
            self.fields_names.add(field_name)
            unique_annotations = extract_unique_annotations(field.annotation)

            if ArFiSettings in unique_annotations:
                raise ArFiSettingsError(
                    f"class {self.settings_class.__class__.__name__}: "
                    f"field `{field_name}` cannot contain ArFiSettings type. "
                    f"Must be only subclass `ArFiSettings`."
                )

            if is_settings(field.annotation):
                self.fields_is_settings.append(field_name)
            if origin_is_union(get_origin(field.annotation)):
                if all([is_settings(arg) for arg in get_args(field.annotation)]):
                    self.fields_is_settings.append(field_name)

            if field_name in self.fields_is_settings:
                if issubclass(type(field.default), ArFiSettings):
                    instance_id = id(field.default)
                    self.fields_defaults[field_name] = instance_id
                if isinstance(field.discriminator, str):
                    discriminator_key = field.discriminator
                    discriminator_value = PydanticUndefined
                    if field.default is not PydanticUndefined:
                        discriminator_value = getattr(field.default, discriminator_key)
                    self.fields_discriminator[field_name] = {discriminator_key: discriminator_value}

            if is_pydantic(field.annotation):
                self.fields_is_pydantic.append(field_name)
                self.pydantic_single_fields[field_name] = self._search_field_keys_info(field.annotation)
            elif origin_is_union(get_origin(field.annotation)):
                if all([is_pydantic(arg) for arg in get_args(field.annotation)]):
                    self.fields_is_pydantic.append(field_name)
                    list_pydantic_aliases_info = []
                    for model in get_args(field.annotation):
                        field_aliases_info_dict = self._search_field_keys_info(model)
                        list_pydantic_aliases_info.append(field_aliases_info_dict)
                    self.pydantic_aliases_info[field_name] = list_pydantic_aliases_info

            if field_name not in self.fields_is_pydantic:
                if allow_json_parse_failure(field):
                    self.allowed_json_parse_failure_fields.add(field_name)

    def _search_field_keys_info(
        self, model: type[BaseModel]
    ) -> tuple[list[dict[str, list]], dict[str, list[str | int]]]:
        """Searches field keys info."""

        list_keys_info = []
        dict_alias_path = dict()
        for field_name, field in model.model_fields.items():
            field_keys = {
                "aliases": [],
                "alias_path": [],
                "keys": [],
            }
            aliases = []
            alias_path = dict()
            if validation_alias := field.validation_alias:
                if isinstance(validation_alias, str):
                    aliases.append(validation_alias)
                    if not self.config.env_case_sensitive:
                        alias_lower = validation_alias.lower()
                        if alias_lower not in aliases:
                            aliases.append(alias_lower)
                else:
                    validation_aliases: list[str | AliasPath] = (
                        field.validation_alias.choices
                        if isinstance(field.validation_alias, AliasChoices)
                        else [field.validation_alias]
                    )
                    for alias in validation_aliases:
                        if isinstance(alias, str):
                            aliases.append(alias)
                            if not self.config.env_case_sensitive:
                                alias_lower = alias.lower()
                                if alias_lower not in aliases:
                                    aliases.append(alias_lower)
                        else:
                            if isinstance(alias, AliasPath):
                                alias = alias.convert_to_aliases()
                            first_arg = alias[0]
                            aliases.append(first_arg)
                            alias_path[first_arg] = alias
                            dict_alias_path[first_arg] = alias
                            if not self.config.env_case_sensitive:
                                alias_lower = first_arg.lower()
                                if alias_lower not in aliases:
                                    aliases.append(alias_lower)
                                    alias_path[alias_lower] = alias
                                    dict_alias_path[alias_lower] = alias
            else:
                aliases.append(field_name)
                if not self.config.env_case_sensitive:
                    alias_lower = field_name.lower()
                    if alias_lower not in aliases:
                        aliases.append(alias_lower)
            field_keys["aliases"] = aliases
            field_keys["alias_path"] = alias_path
            if is_pydantic(field.annotation):
                field_aliases_info_dict = self._search_field_keys_info(field.annotation)
                field_keys["keys"] = field_aliases_info_dict["list_keys_info"]
            list_keys_info.append(field_keys)

        list_nested_keys = self._search_field_nested_keys(list_keys_info)
        field_aliases_info_dict = dict(
            model_name=model.__name__,
            list_keys_info=list_keys_info,
            dict_alias_path=dict_alias_path,
            list_nested_keys=list_nested_keys,
        )
        return field_aliases_info_dict

    def _search_field_nested_keys(self, list_keys_info: list[dict[str, list]]) -> list[str]:
        """Searches field nested keys."""

        list_nested_keys = []
        env_nested_delimiter = self.config.env_nested_delimiter
        if env_nested_delimiter:
            for field_keys in list_keys_info:
                aliases = field_keys["aliases"]
                alias_path = field_keys["alias_path"]
                nested_aliases = []
                if field_keys["keys"]:
                    nested_aliases = self._search_field_nested_keys(field_keys["keys"])
                if nested_aliases:
                    list_aliases = [aliases, nested_aliases]
                    list_aliases_str = []
                    for combo in itertools.product(*list_aliases):
                        if combo[0] in alias_path:
                            continue
                        else:
                            combo_str = f"{env_nested_delimiter}".join(combo)
                        list_aliases_str.append(combo_str)
                    list_aliases_str.extend(aliases)
                    list_nested_keys.extend(list_aliases_str)
                else:
                    list_nested_keys.extend(aliases)
        return list_nested_keys

    @classmethod
    def _is_exist_main_handler(cls, name: str) -> bool:
        """Check is exist main handler."""

        name = cls._normalize_handler_name(name, "main")
        return hasattr(cls, name)

    @classmethod
    def _is_exist_ordered_settings_handler(cls, name: str) -> bool:
        """Check is exist ordered_settings handler."""

        name = cls._normalize_handler_name(name, "ordered_settings")
        return hasattr(cls, name)

    @classmethod
    def _is_exist_ext_handler(cls, name: str) -> bool:
        """Check is exist extensions handler."""

        name = cls._normalize_handler_name(name, "ext")
        return hasattr(cls, name)

    @classmethod
    def _validate_main_handler(cls, handler: str) -> str:
        """Validate name and check exist main handler."""

        if not handler or not isinstance(handler, str):
            raise ArFiSettingsError(f"{cls.__name__}: attribute `handler` must be not empty string")
        if not handler.endswith("_main_handler"):
            raise ArFiSettingsError(f"{cls.__name__}: attribute `handler` must end with `_main_handler`")
        if not cls._is_exist_main_handler(handler):
            raise ArFiSettingsError(f"{cls.__name__}: `{handler}` is not defined")
        return handler

    @classmethod
    def _validate_ordered_settings_handlers(cls, ordered_settings: list[str], settings_class: type) -> list[str]:
        """Returns a list of valid ordered_settings handlers."""

        if not ordered_settings:
            raise ArFiSettingsError(f"{settings_class.__class__.__name__}: `ordered_settings` is empty")
        valid_ordered_settings = []
        for handler_name in ordered_settings:
            handler_name = cls._normalize_handler_name(handler_name, "ordered_settings")
            if not cls._is_exist_ordered_settings_handler(handler_name):
                raise ArFiSettingsError(f"{cls.__name__}: `{handler_name}` is not defined")
            valid_ordered_settings.append(handler_name)
        if "init_kwargs_ordered_settings_handler" not in valid_ordered_settings:
            warnings.warn(
                f"\033[33m\n"
                f"For instance of {settings_class.__class__.__name__}():\n"
                f"Reverse inheritance is disabled because `init_kwargs` is missing in `ordered_settings`"
                f"\033[0m",
                category=Warning,
                stacklevel=4,
            )
        return valid_ordered_settings

    def _validate_conf_custom_ext_handler(
        self,
        conf_custom_ext_handler: str | dict[str, str] | None,
    ) -> dict[str, str]:
        """Validate `conf_custom_ext_handler` and return valid dict."""

        valid_conf_custom_ext_handler = {}
        error_message = (
            f"{self.settings_class.__class__.__name__}: `conf_ext` contains empty extension, "
            f"but `conf_custom_ext_handler` is not defined"
        )
        if conf_custom_ext_handler is None or conf_custom_ext_handler == "":
            if "" in self.config.conf_ext:
                raise ArFiSettingsError(error_message)
            return valid_conf_custom_ext_handler
        if isinstance(conf_custom_ext_handler, str):
            conf_custom_ext_handler = {"": conf_custom_ext_handler}
        if not isinstance(conf_custom_ext_handler, dict):
            raise ArFiSettingsError(
                f"{self.settings_class.__class__.__name__}: `conf_custom_ext_handler` "
                f"must be str, dict[str,str] or None. Current type: {type(conf_custom_ext_handler)}"
            )
        for ext, handler_name in conf_custom_ext_handler.items():
            assert isinstance(ext, str), (
                f"{self.settings_class.__class__.__name__}: Not valid type 'ext' in `conf_custom_ext_handler`. "
                f"Must be str. Current type: {type(ext)}"
            )
            handler_name = handler_name.lstrip(".")
            if ext and isinstance(ext, str):
                ext = ext.lstrip(".")
            handler_name = self._normalize_handler_name(handler_name, "ext")
            if not self._is_exist_ext_handler(handler_name):
                raise ArFiSettingsError(f"{self.__class__.__name__}: `{handler_name}` is not defined")
            valid_conf_custom_ext_handler[ext] = handler_name

        if "" in self.config.conf_ext and not valid_conf_custom_ext_handler.get("", False):
            raise ArFiSettingsError(error_message)
        self.config.conf_custom_ext_handler = valid_conf_custom_ext_handler
        return valid_conf_custom_ext_handler

    def _validate_conf_ext(self, conf_ext: str | list[str]) -> list[str]:
        """Validate conf_ext."""

        valid_conf_ext = []
        error_message = (
            f"{self.settings_class.__class__.__name__}: `conf_ext` "
            f"must be str, list[str] or empty string. Current: {conf_ext}"
        )
        if isinstance(conf_ext, str):
            conf_ext = conf_ext.lstrip(".")
            if not conf_ext:
                return [""]
            return [conf_ext]
        elif isinstance(conf_ext, list):
            for ext in conf_ext:
                if isinstance(ext, str):
                    ext = ext.lstrip(".")
                    if ext not in valid_conf_ext:
                        valid_conf_ext.append(ext)
                else:
                    raise ArFiSettingsError(error_message)
        else:
            raise ArFiSettingsError(error_message)
        return valid_conf_ext

    def _get_main_handler(self, name: str) -> Callable:
        name = self._normalize_handler_name(name, "main")
        if not self._is_exist_main_handler(name):
            raise ArFiSettingsError(f"{self.__class__.__name__}: `{name}` is not defined")
        return self.__getattribute__(name)

    def _get_ordered_settings_handler(self, name: str) -> Callable:
        name = self._normalize_handler_name(name, "ordered_settings")
        if not self._is_exist_ordered_settings_handler(name):
            raise ArFiSettingsError(f"{self.__class__.__name__}: `{name}` is not defined")
        return self.__getattribute__(name)

    def _find_ext_handler(self, file_path: Path) -> (Callable | None, Path):
        """Find and return ext handler."""

        source_file_path = file_path
        if ext := file_path.suffix:
            ext = ext.lstrip(".")
            ext = self.conf_custom_ext_handler.get(ext, ext)
            ext = ext.lstrip(".")
            handler = self._get_ext_handler(ext)
            return handler, file_path

        assert isinstance(self.config.conf_ext, list) and self.config.conf_ext

        for ext in self.config.conf_ext:
            if not ext:
                file_path = file_path.with_suffix("")
                if self.reader_class.is_exist_file(file_path):
                    name_custom_ext_handler = self.conf_custom_ext_handler.get("")
                    if name_custom_ext_handler is None:
                        raise ArFiSettingsError(f"Missing non extension handler for file: `{file_path.as_posix()}`")
                    handler = self._get_ext_handler(name_custom_ext_handler)
                    return handler, file_path
            else:
                ext = ext.lstrip(".")
                file_ext = f".{ext}"
                file_path = file_path.with_suffix(file_ext)
                if self.reader_class.is_exist_file(file_path):
                    name_custom_ext_handler = self.conf_custom_ext_handler.get(ext)
                    if name_custom_ext_handler:
                        ext = name_custom_ext_handler.lstrip(".")
                    handler = self._get_ext_handler(ext)
                    return handler, file_path
        return None, source_file_path

    def _get_ext_handler(self, name: str) -> Callable:
        name = self._normalize_handler_name(name, "ext")
        if not self._is_exist_ext_handler(name):
            raise ArFiSettingsError(f"{self.__class__.__name__}: `{name}` is not defined")
        return self.__getattribute__(name)

    @staticmethod
    def _normalize_handler_name(name: str, suffix: str) -> str:
        """Normalize handler name."""

        try:
            assert name and isinstance(name, str), "Handler `name` must be not empty string"
            assert not name.startswith("_"), "Handler `name` must not start with `_`"
        except AssertionError as e:
            raise ArFiSettingsError(e) from e
        if not suffix.endswith("_handler"):
            suffix = f"{suffix}_handler"
        suffix = suffix.lstrip("_")
        try:
            assert suffix in ["main_handler", "ordered_settings_handler", "ext_handler"], "Unknown handler suffix"
        except AssertionError as e:
            raise ArFiSettingsError(e) from e
        if not name.endswith(suffix):
            name = f"{name}_{suffix}"
        return name

    def toml_ext_handler(self, file_path: PathType) -> dict[str, Any]:
        """Handles settings from .toml file."""

        reader = self.reader_class(
            file_path=file_path,
            file_encoding=self.config.conf_file_encoding,
            ignore_missing=self.config.conf_ignore_missing,
        )
        data = reader.read()
        data["__case_sensitive"] = self.config.conf_case_sensitive
        return data

    def yaml_ext_handler(self, file_path: PathType) -> dict[str, Any]:
        """Handles settings from .yaml file."""

        reader = self.reader_class(
            file_path=file_path,
            file_encoding=self.config.conf_file_encoding,
            ignore_missing=self.config.conf_ignore_missing,
        )
        data = reader.read() or {}
        data["__case_sensitive"] = self.config.conf_case_sensitive
        return data

    def yml_ext_handler(self, file_path: PathType) -> dict[str, Any]:
        """Handles settings from .yml file."""

        reader = self.reader_class(
            file_path=file_path,
            file_encoding=self.config.conf_file_encoding,
            ignore_missing=self.config.conf_ignore_missing,
        )
        data = reader.read() or {}
        data["__case_sensitive"] = self.config.conf_case_sensitive
        return data

    def json_ext_handler(self, file_path: PathType) -> dict[str, Any]:
        """Handles settings from .json file."""

        reader = self.reader_class(
            file_path=file_path,
            file_encoding=self.config.conf_file_encoding,
            ignore_missing=self.config.conf_ignore_missing,
        )
        data = reader.read()
        data["__case_sensitive"] = self.config.conf_case_sensitive
        return data

    def cli_ordered_settings_handler(self) -> dict[str, Any]:
        """Handles settings from CLI."""

        data: dict[str, Any] = {}
        if self.config.cli:
            reader = self.reader_class(is_cli=True)
            data = reader.read()
            data["__case_sensitive"] = self.config.case_sensitive
        return data

    def init_kwargs_ordered_settings_handler(self) -> dict[str, Any]:
        """Handles settings from init_kwargs."""

        return self.init_kwargs

    def env_ordered_settings_handler(self) -> dict[str, Any]:
        """Handles settings from environment."""

        data: dict[str, Any] = {}
        reader = self.reader_class(
            is_env=True,
        )
        reader_data = reader.read()
        reader_data["__case_sensitive"] = self.config.env_case_sensitive
        data = self._convert_env_data_by_fields_names(data=reader_data, handler="env")
        return data

    def env_file_ordered_settings_handler(self) -> dict[str, Any]:
        """Handles settings from .env file."""

        data: dict[str, Any] = {}
        for file_path in self.config.env_path:
            reader = self.reader_class(
                file_path=file_path,
                is_env_file=True,
                file_encoding=self.config.env_file_encoding,
                ignore_missing=self.config.env_ignore_missing,
            )
            reader_data = reader.read()
            reader_data["__case_sensitive"] = self.config.env_case_sensitive
            env_file_data = self._convert_env_data_by_fields_names(
                data=reader_data,
                handler="env_file",
                env_file=file_path,
            )
            data = deep_update(data, env_file_data)

        return data

    def _convert_env_data_by_fields_names(
        self,
        data: dict[str, Any],
        handler: Literal["env_file", "env"],  # for feature debug
        env_file: Path | None = None,  # for feature debug
    ) -> dict[str, Any]:
        """Converts data from environment to data with fields names."""

        valid_data: dict[str, Any] = dict()
        fields_set = set()
        case_sensitive = data.get("__case_sensitive", self.config.case_sensitive)
        lower_data = dict()
        if case_sensitive:
            dict_fields_alias = self.field_aliases
            dict_fields_alias_path = self.fields_alias_path
        else:
            dict_fields_alias = self.field_aliases_not_case_sensitive
            dict_fields_alias_path = self.fields_alias_path_not_case_sensitive
            for key, value in data.items():
                if isinstance(key, str):
                    key = key.lower()
                if key not in lower_data:
                    lower_data[key] = value
        env_prefix = self.config.env_prefix
        env_nested_delimiter = self.config.env_nested_delimiter
        handler_tree = self.settings_class._handler_tree

        source_parents_prefixis = [""]
        if env_nested_delimiter and handler_tree:
            source_parents_prefixis = []

            for combo in itertools.product(*handler_tree):
                combo_str = f"{env_nested_delimiter}".join(combo)
                parent_pref = f"{combo_str}{env_nested_delimiter}"
                source_parents_prefixis.append(parent_pref)
        source_parents_prefixis = source_parents_prefixis or [""]
        parents_prefixis = []
        for parent_prefix in source_parents_prefixis:
            parents_prefixis.append(f"{parent_prefix}{env_prefix}")

        for field_name, aliases in dict_fields_alias.items():
            fields_alias_path = dict_fields_alias_path.get(field_name)

            found_value = self._search_exact_env_value(
                field_name=field_name,
                aliases=aliases,
                parents_prefixis=parents_prefixis,
                env_nested_delimiter=env_nested_delimiter,
                data=data,
                lower_data=lower_data,
                fields_set=fields_set,
                fields_alias_path=fields_alias_path,
            )

            # TODO: Combine fields_is_settings и fields_is_pydantic. Take out only the discriminator !!!
            if field_name in self.fields_is_settings:
                found_value = self._search_env_value_arfi_settings_fields(
                    field_name=field_name,
                    aliases=aliases,
                    parents_prefixis=parents_prefixis,
                    env_nested_delimiter=env_nested_delimiter,
                    found_value=found_value,
                    data=data,
                    lower_data=lower_data,
                )
            elif field_name in self.fields_is_pydantic:
                if isinstance(found_value, str):
                    found_value = json.loads(found_value)
                if field_name in self.pydantic_single_fields:
                    found_value = self._search_env_value_pydantic_single_field(
                        field_name=field_name,
                        aliases=aliases,
                        parents_prefixis=parents_prefixis,
                        env_nested_delimiter=env_nested_delimiter,
                        found_value=found_value,
                        data=data,
                    )
            else:
                if found_value is not PydanticUndefined:
                    try:
                        found_value = json.loads(found_value)
                    except (ValueError, json.decoder.JSONDecodeError) as e:
                        if field_name in self.allowed_json_parse_failure_fields:
                            pass
                        else:
                            raise e

            if found_value is not PydanticUndefined:
                valid_data[field_name] = found_value
        return valid_data

    def _search_exact_env_value(
        self,
        field_name: str,
        aliases: list[str],
        parents_prefixis: list[str],
        env_nested_delimiter: str,
        data: dict[str, Any],
        lower_data: dict[str, Any],
        fields_set: set[str],
        fields_alias_path: dict[str, AliasPath | list[str | int]],
    ) -> Union[dict[str, Any], PydanticUndefined]:
        """Searches exact value in environment."""

        found_value = PydanticUndefined
        for prefix in parents_prefixis:
            for alias in aliases:
                search_alias = f"{prefix}{alias}"
                if field_name in fields_set:
                    break

                value = data.get(search_alias, PydanticUndefined)
                if not self.config.env_case_sensitive and value is PydanticUndefined:
                    if prefix:
                        for k, v in data.items():
                            if k.startswith(prefix):
                                if k.lower() == search_alias.lower():
                                    value = v

                    if value is PydanticUndefined:
                        search_alias = f"{prefix}{alias.lower()}"
                        value = lower_data.get(search_alias, PydanticUndefined)

                if value is not PydanticUndefined:
                    if fields_alias_path:
                        if alias_path := fields_alias_path.get(alias):
                            value = json.loads(value)
                            value_dict = {alias: value}
                            field_value = search_dict_for_path(
                                alias_path,
                                value_dict,
                                self.config.env_case_sensitive,
                            )
                            found_value = field_value
                            fields_set.add(field_name)
                        else:
                            found_value = value
                            fields_set.add(field_name)
                    else:
                        found_value = value
                        fields_set.add(field_name)
                    break
            if value is not PydanticUndefined:
                break
        return found_value

    def _search_env_value_arfi_settings_fields(
        self,
        field_name: str,
        aliases: list[str],
        parents_prefixis: list[str],
        env_nested_delimiter: str,
        found_value: Union[dict, PydanticUndefined],
        data: dict[str, Any],
        lower_data: dict[str, Any],
    ) -> Union[dict[str, Any], PydanticUndefined]:
        """Searches value in arfi_settings fields."""

        if isinstance(found_value, str):
            found_value = json.loads(found_value)
            assert isinstance(found_value, dict)
        if field_name in self.pydantic_single_fields:
            # convert exact value
            if found_value is not PydanticUndefined:
                assert isinstance(found_value, dict)
                found_value = self._convert_env_found_value_key(
                    list_keys_info=self.pydantic_single_fields[field_name]["list_keys_info"],
                    found_value=found_value,
                    field_name=field_name,
                )
            # search nested key-value
            found_value = self._search_env_value_pydantic_single_field(
                field_name=field_name,
                aliases=aliases,
                parents_prefixis=parents_prefixis,
                env_nested_delimiter=env_nested_delimiter,
                found_value=found_value,
                data=data,
            )
        else:
            list_pydantic_aliases_info = self.pydantic_aliases_info[field_name]
            for field_aliases_info_dict in reversed(list_pydantic_aliases_info):
                # Search velue of nested key
                nested_key_dict = self._search_nested_key_value(
                    field_name=field_name,
                    aliases=aliases,
                    parents_prefixis=parents_prefixis,
                    env_nested_delimiter=env_nested_delimiter,
                    found_value=found_value,
                    data=data,
                    field_aliases_info_dict=field_aliases_info_dict,
                )
                if nested_key_dict:
                    if found_value is PydanticUndefined:
                        found_value = dict()
                    found_value = deep_update(found_value, nested_key_dict)

        # Search discriminator
        discriminator = self.fields_discriminator.get(field_name)
        if discriminator:
            if found_value is PydanticUndefined:
                found_value = dict()
            discriminator_key = list(discriminator.keys())[0]
            alias = aliases[0]
            for PREFIX in parents_prefixis:
                if env_nested_delimiter:
                    search_alias = f"{PREFIX}{alias}{env_nested_delimiter}{discriminator_key}"
                else:
                    search_alias = f"{PREFIX}{alias}_{discriminator_key}"

                value = data.get(search_alias, PydanticUndefined)
                if not self.config.env_case_sensitive and value is PydanticUndefined:
                    value = lower_data.get(search_alias.lower(), PydanticUndefined)
                if value is not PydanticUndefined:
                    found_value[discriminator_key] = value
                    break
        return found_value

    def _search_env_value_pydantic_single_field(
        self,
        field_name: str,
        aliases: list[str],
        parents_prefixis: list[str],
        env_nested_delimiter: str,
        found_value: Union[dict, PydanticUndefined],
        data: dict[str, Any],
    ) -> Union[dict[str, Any], PydanticUndefined]:
        """Searches value in pydantic single fields."""

        if found_value is not PydanticUndefined:
            assert isinstance(found_value, dict)
            found_value = self._convert_env_found_value_key(
                list_keys_info=self.pydantic_single_fields[field_name]["list_keys_info"],
                found_value=found_value,
                field_name=field_name,
            )

        # Search velue of nested key
        nested_key_dict = self._search_nested_key_value(
            field_name=field_name,
            aliases=aliases,
            parents_prefixis=parents_prefixis,
            env_nested_delimiter=env_nested_delimiter,
            found_value=found_value,
            data=data,
            field_aliases_info_dict=self.pydantic_single_fields[field_name],
        )

        if nested_key_dict:
            if found_value is PydanticUndefined:
                found_value = dict()
            found_value = deep_update(found_value, nested_key_dict)
        return found_value

    def _search_nested_key_value(
        self,
        field_name: str,
        aliases: list[str],
        parents_prefixis: list[str],
        env_nested_delimiter: str,
        found_value: Union[dict, PydanticUndefined],
        data: dict[str, Any],
        field_aliases_info_dict: dict[str, Any],
    ) -> dict[str, Any]:
        """Search nested key value in environment variable."""

        list_keys_info = field_aliases_info_dict["list_keys_info"]
        list_nested_keys = field_aliases_info_dict["list_nested_keys"]
        dict_alias_path = field_aliases_info_dict["dict_alias_path"]
        nested_key_lower = {i.lower() for i in list_nested_keys}

        nested_key_dict = dict()
        for prefix in reversed(parents_prefixis):
            for alias in reversed(aliases):
                start_key_prefix = f"{prefix}{alias}{env_nested_delimiter}"
                env_find_data = dict()
                for env_key, env_value in data.items():
                    if env_key.startswith(start_key_prefix):
                        env_nested_key = env_key[len(start_key_prefix) :]
                        if env_nested_key.lower() in nested_key_lower:
                            env_find_data[env_key] = env_value
                    elif not self.config.case_sensitive:
                        env_key_lower = env_key.lower()
                        if env_key_lower.startswith(start_key_prefix):
                            env_nested_key_lower = env_key_lower[len(start_key_prefix) :]
                            if env_nested_key_lower in nested_key_lower:
                                env_find_data[env_key] = env_value
                if env_find_data:
                    if self.config.case_sensitive:
                        for nested_key in list_nested_keys:
                            alias_path_key = nested_key.split(env_nested_delimiter)[0]
                            path_alias = dict_alias_path.get(alias_path_key, [])
                            for env_key, env_value in env_find_data.items():
                                search_key = f"{start_key_prefix}{nested_key}"
                                if env_key.startswith(search_key):
                                    key_path = nested_key.split(nested_key)
                                    found_dict = create_dict_for_path(key_path, env_value)
                                    found_dict = self._convert_env_found_value_key(
                                        list_keys_info=list_keys_info,
                                        found_value=found_dict,
                                        field_name=field_name,
                                        is_alias_path=path_alias,
                                    )
                                    nested_key_dict = deep_update(nested_key_dict, found_dict)
                    else:
                        for nested_key in reversed(list_nested_keys):
                            is_alias_path = False
                            alias_path_key = nested_key.split(env_nested_delimiter)[0]
                            if dict_alias_path.get(alias_path_key):
                                is_alias_path = True
                            exact_match_dict = dict()
                            middle_match_dict = dict()
                            lower_match_dict = dict()
                            for env_key, env_value in env_find_data.items():
                                env_nested_key = env_key[len(start_key_prefix) :]

                                if env_key.startswith(start_key_prefix):
                                    if nested_key == env_nested_key:
                                        key_path = env_nested_key.split(env_nested_delimiter)
                                        found_dict = create_dict_for_path(key_path, env_value)
                                        found_dict = self._convert_env_found_value_key(
                                            list_keys_info=list_keys_info,
                                            found_value=found_dict,
                                            field_name=field_name,
                                            is_alias_path=is_alias_path,
                                        )
                                        exact_match_dict = deep_update(exact_match_dict, found_dict)
                                    elif nested_key.lower() == env_nested_key.lower():
                                        key_path = env_nested_key.split(env_nested_delimiter)
                                        found_dict = create_dict_for_path(key_path, env_value)
                                        found_dict = self._convert_env_found_value_key(
                                            list_keys_info=list_keys_info,
                                            found_value=found_dict,
                                            field_name=field_name,
                                            is_alias_path=is_alias_path,
                                        )
                                        middle_match_dict = deep_update(middle_match_dict, found_dict)
                                elif env_key.lower().startswith(start_key_prefix):
                                    if nested_key == env_nested_key:
                                        key_path = env_nested_key.split(env_nested_delimiter)
                                        found_dict = create_dict_for_path(key_path, env_value)
                                        found_dict = self._convert_env_found_value_key(
                                            list_keys_info=list_keys_info,
                                            found_value=found_dict,
                                            field_name=field_name,
                                            is_alias_path=is_alias_path,
                                        )
                                        middle_match_dict = deep_update(middle_match_dict, found_dict)
                                    elif nested_key.lower() == env_nested_key.lower():
                                        key_path = env_nested_key.split(env_nested_delimiter)
                                        found_dict = create_dict_for_path(key_path, env_value)
                                        found_dict = self._convert_env_found_value_key(
                                            list_keys_info=list_keys_info,
                                            found_value=found_dict,
                                            field_name=field_name,
                                            is_alias_path=is_alias_path,
                                        )
                                        lower_match_dict = deep_update(lower_match_dict, found_dict)
                                else:
                                    raise ValueError(f"Unresolved nested key: {nested_key}")

                            nested_key_dict = deep_update(nested_key_dict, lower_match_dict)
                            nested_key_dict = deep_update(nested_key_dict, middle_match_dict)
                            nested_key_dict = deep_update(nested_key_dict, exact_match_dict)

        return nested_key_dict

    def _convert_env_found_value_key(
        self,
        list_keys_info: list[dict[str, list]],
        found_value: str | dict[str, Any],
        field_name: str = "",
        is_alias_path: bool = False,
    ) -> dict[str, Any]:
        """Convert found environment variable value to model field names."""

        valid_data: dict[str, Any] = dict()
        found_value_lower = dict()

        if isinstance(found_value, str):
            found_value = json.loads(found_value)

        assert isinstance(found_value, dict)

        if not self.config.env_case_sensitive:
            for k, v in found_value.items():
                k = k.lower()
                if k not in found_value_lower:
                    found_value_lower[k] = v
        for field_keys in list_keys_info:
            aliases = field_keys["aliases"]
            alias_path = field_keys["alias_path"]
            keys = field_keys["keys"]
            main_alias = aliases[0]
            result_value = PydanticUndefined

            find_dict = {}
            for alias in reversed(aliases):
                result_value = found_value.get(alias, PydanticUndefined)
                if result_value is PydanticUndefined and not self.config.env_case_sensitive:
                    alias_lower = alias.lower()
                    result_value = found_value_lower.get(alias_lower, PydanticUndefined)

                if alias in alias_path:
                    if isinstance(result_value, str):
                        result_value = json.loads(result_value)
                if result_value is not PydanticUndefined:
                    if keys and not is_alias_path and alias not in alias_path:
                        result_value = self._convert_env_found_value_key(
                            list_keys_info=keys,
                            found_value=result_value,
                            field_name=alias,
                        )
                if result_value is not PydanticUndefined:
                    if alias in alias_path:
                        if isinstance(result_value, str):
                            result_value = json.loads(result_value)
                        path_alias = alias_path[alias]
                        search_value = search_dict_for_path(
                            path_alias, {alias: result_value}, self.config.env_case_sensitive
                        )
                        if search_value is not PydanticUndefined:
                            result_value = create_dict_for_path(path_alias, search_value)
                        find_dict = deep_update(find_dict, result_value)
                    else:
                        if main_alias in alias_path:
                            path_alias = alias_path[main_alias]
                            result_value = create_dict_for_path(path_alias, result_value)
                            find_dict = deep_update(find_dict, result_value)
                        else:
                            find_dict[main_alias] = result_value
            if find_dict:
                valid_data = deep_update(valid_data, find_dict)

        return valid_data

    def secrets_ordered_settings_handler(self) -> dict[str, Any]:
        """Handles settings from secrets directory."""

        data: dict[str, Any] = {}
        if self.config.secrets_dir is None:
            return data
        data["__case_sensitive"] = self.config.case_sensitive
        if isinstance(self.config.secrets_dir, str):
            self.config.secrets_dir = Path(self.config.secrets_dir)

        assert isinstance(self.config.secrets_dir, Path)

        if not self.config.secrets_dir.is_dir():
            if self.config.ignore_missing:
                return data
            raise ArFiSettingsError(f"Missing secrets directory: `{self.config.secrets_dir.as_posix()}`")

        all_files = self.config.secrets_dir.rglob("*")
        for file in all_files:
            if not file.is_file():
                continue
            if file.suffix:
                continue
            file_name = file.stem
            try:
                data[file_name] = file.read_text(encoding=self.config.encoding).strip()
            except UnicodeDecodeError as e:
                raise ArFiSettingsError(f"Error reading file: `{file.as_posix()}`") from e

        return data

    def conf_file_ordered_settings_handler(self, mode: str | None = None) -> dict[str, Any]:
        """Handles settings from config file."""

        data: dict[str, Any] = {}
        for file_path in self.config.conf_path:
            if not file_path.is_absolute():
                if self.settings_class.BASE_DIR is not None:
                    base_dir = Path(self.settings_class.BASE_DIR).resolve()
                    file_path = Path(base_dir, file_path).resolve()
            if mode:
                file_path = file_path.parent / mode
            handler, file_path = self._find_ext_handler(file_path)
            if handler is None:
                if not self.config.conf_ignore_missing:
                    raise ArFiSettingsError(f"Missing file: `{file_path.as_posix()}`")
                continue
            data = deep_update(data, handler(file_path=file_path))

        return data

    @abstractmethod
    def default_main_handler(self) -> dict[str, Any]:
        """Main handler by default."""


class ArFiHandler(ArFiBaseHandler):
    """Handles source settings."""

    def default_main_handler(self) -> dict[str, Any]:
        """Main handler."""

        data: dict[str, Any] = {}
        mode = self.data.get("MODE", None)
        for ord_handler in reversed(self.ordered_settings):
            handler = self._get_ordered_settings_handler(ord_handler)
            handler_data = handler()
            current_mode = handler_data.get("MODE", None)
            if current_mode:
                mode = current_mode
            data[ord_handler] = handler_data

        mode_field = self.settings_class.model_fields.get("MODE")
        class_mode = mode_field.default
        mode = mode or class_mode
        if mode:
            if "conf_file_ordered_settings_handler" in self.ordered_settings:
                handler_data = self.conf_file_ordered_settings_handler(mode=mode)
                data["conf_file_ordered_settings_handler"] = deep_update(
                    data["conf_file_ordered_settings_handler"], handler_data
                )

        for ord_handler in reversed(self.ordered_settings):
            self.data = deep_update(self.data, data[ord_handler])
        return self.data



import pytest

from arfi_settings import ArFiSettings, ArFiSettingsError


@pytest.mark.field_type
def test_legal_field_type(cwd_to_tmp, path_base_dir):
    class CustomSettings(ArFiSettings):
        pass

    class AppSettings(ArFiSettings):
        pass

    class AppConfig(ArFiSettings):
        app: AppSettings = AppSettings()

    config = AppConfig()
    assert config

    class AppConfig(ArFiSettings):
        app: CustomSettings | AppSettings = AppSettings()

    config = AppConfig()
    assert config

    class AppConfig(ArFiSettings):
        app: list[AppSettings | CustomSettings] = [
            AppSettings(),
            CustomSettings(),
        ]

    config = AppConfig()
    assert config


@pytest.mark.field_type
def test_illegal_field_type(cwd_to_tmp, path_base_dir):
    class AppSettings(ArFiSettings):
        pass

    with pytest.raises(ArFiSettingsError) as excinfo:

        class AppConfig(ArFiSettings):
            app: ArFiSettings = AppSettings()

        _ = AppConfig()
    assert "cannot contain ArFiSettings type" in str(excinfo)

    with pytest.raises(ArFiSettingsError) as excinfo:

        class AppConfig(ArFiSettings):
            app: ArFiSettings | None = AppSettings()

        _ = AppConfig()
    assert "cannot contain ArFiSettings type" in str(excinfo)

    with pytest.raises(ArFiSettingsError) as excinfo:

        class AppConfig(ArFiSettings):
            app: list[ArFiSettings] = [
                AppSettings(),
            ]

        _ = AppConfig()
    assert "cannot contain ArFiSettings type" in str(excinfo)

    with pytest.raises(ArFiSettingsError) as excinfo:

        class AppConfig(ArFiSettings):
            app: dict[str, ArFiSettings] = dict(asd=AppSettings())

        _ = AppConfig()
    assert "cannot contain ArFiSettings type" in str(excinfo)

    with pytest.raises(ArFiSettingsError) as excinfo:

        class AppConfig(ArFiSettings):
            app: AppSettings | tuple[dict[str, ArFiSettings], str] | str = AppSettings()

        _ = AppConfig()
    assert "cannot contain ArFiSettings type" in str(excinfo)



import pytest
from pydantic.aliases import AliasPath
from pydantic_core import PydanticUndefined

from arfi_settings.utils import create_dict_for_path, search_dict_for_path


@pytest.mark.utils
def test_create_dict_for_path():
    path = AliasPath("foo", "bar", "baz")
    res = create_dict_for_path(path)
    assert res == {"foo": {"bar": "baz"}}

    res = create_dict_for_path(path, "asd")
    assert res == {"foo": {"bar": {"baz": "asd"}}}

    path = ["foo", "bar", "baz"]
    res = create_dict_for_path(path)
    assert res == {"foo": {"bar": "baz"}}

    res = create_dict_for_path(path, "asd")
    assert res == {"foo": {"bar": {"baz": "asd"}}}


# @pytest.mark.current
@pytest.mark.utils
def test_search_dict_for_path():
    search_dict = {"foo": {"bar": {"baz": "asd"}}}
    path = AliasPath("foo", "bar", "baz")
    value = search_dict_for_path(path, search_dict)
    assert value == "asd"

    path = ["foo", "bar", "baz"]
    value = search_dict_for_path(path, search_dict)
    assert value == "asd"

    path = ["params", 1]
    search_dict = {"params": ["foo", "bar"]}
    value = search_dict_for_path(path, search_dict)
    assert value == "bar"

    value = search_dict_for_path(["params", 8], search_dict)
    assert value is PydanticUndefined

    value = search_dict_for_path(["params", 8], search_dict, case_sensitive=False)
    assert value is PydanticUndefined

    value = search_dict_for_path(["params", 8], "search_dict")
    assert value is PydanticUndefined

    value = search_dict_for_path(["params", 8], "search_dict", case_sensitive=False)
    assert value is PydanticUndefined

    with pytest.raises(AssertionError) as excinfo:
        search_dict_for_path("params", search_dict)
    assert "`alias_path` must be a AliasPath or list" in str(excinfo.value)

    with pytest.raises(AssertionError) as excinfo:
        search_dict_for_path("params", search_dict, case_sensitive=False)

    assert "`alias_path` must be a AliasPath or list" in str(excinfo.value)

    value = search_dict_for_path(AliasPath("params"), search_dict, case_sensitive=False)
    assert value == ["foo", "bar"]

    value = search_dict_for_path(["params"], search_dict, case_sensitive=False)
    assert value == ["foo", "bar"]

    search_dict = {"foo": {"BAR": {"baz": "asd"}}}
    path = AliasPath("FOO", "Bar", "baZ")
    value = search_dict_for_path(path, search_dict, case_sensitive=False)
    assert value == "asd"

    search_dict = {"FOO": {"BAR": {"BAZ": "ASD"}}}
    path = AliasPath("foo", "bar", "baz")
    value = search_dict_for_path(path, search_dict, case_sensitive=False)
    assert value == "ASD"

    search_dict = {"foo": {"BAR": {3: "asd"}}}
    path = AliasPath("FOO", "Bar", 3)
    value = search_dict_for_path(path, search_dict, case_sensitive=False)
    assert value == "asd"

    search_dict = {"foo": {3: {"BAR": "asd"}}}
    path = AliasPath("FOO", 3, "Bar")
    value = search_dict_for_path(path, search_dict, case_sensitive=False)
    assert value == "asd"

    search_dict = {3: {2: {1: []}}}
    path = AliasPath(3, 2, 1)
    value = search_dict_for_path(path, search_dict, case_sensitive=False)
    assert value == []

    search_dict = {3: {2: {(1, 0): []}}}
    path = AliasPath(3, 2, (1, 0))
    value = search_dict_for_path(path, search_dict, case_sensitive=False)
    assert value == []



from pathlib import PosixPath

import pydantic
import pytest
from pydantic import AliasChoices, AliasPath, Field

from arfi_settings import (
    ArFiSettings,
    ArFiSettingsError,
    SettingsConfigDict,
)


# @pytest.mark.current
@pytest.mark.secret
def test_non_existing_secrets_dir(cwd_to_tmp):
    class AppConfig(ArFiSettings):
        ordered_settings = [
            "secrets",
        ]
        model_config = SettingsConfigDict(
            secrets_dir="non_existing_secret_dir",
        )

    config = AppConfig()
    assert config.settings_config.secrets_dir == PosixPath("non_existing_secret_dir")

    with pytest.raises(ArFiSettingsError) as excinfo:

        class AppConfig(ArFiSettings):
            ordered_settings = [
                "secrets",
            ]
            model_config = SettingsConfigDict(
                secrets_dir="non_existing_secrets_dir",
                ignore_missing=False,
            )

        _ = AppConfig()
    assert "Missing secrets directory" in str(excinfo.value)


# @pytest.mark.current
@pytest.mark.secret
def test_empty_secrets_dir(secrets_dir):
    class AppConfig(ArFiSettings):
        ordered_settings = [
            "secrets",
        ]
        model_config = SettingsConfigDict(
            secrets_dir="secrets",
            ignore_missing=False,
        )

    config = AppConfig()
    assert config.settings_config.secrets_dir == PosixPath("secrets")


# @pytest.mark.current
@pytest.mark.secret
def test_empty_data_secrets_dir(empty_secret_path_config_file):
    class AppConfig(ArFiSettings):
        path_config_file: str
        ordered_settings = [
            "secrets",
        ]
        model_config = SettingsConfigDict(
            secrets_dir="secrets",
            ignore_missing=False,
        )

    config = AppConfig()
    assert config.settings_config.secrets_dir == PosixPath("secrets")
    assert config.path_config_file == ""


# @pytest.mark.current
@pytest.mark.secret
def test_simple_data_secrets_dir(simple_data_secret_path_config_file):
    class AppConfig(ArFiSettings):
        path_config_file: str
        ordered_settings = [
            "secrets",
        ]
        model_config = SettingsConfigDict(
            secrets_dir="secrets",
            ignore_missing=False,
        )

    config = AppConfig()
    assert config.settings_config.secrets_dir == PosixPath("secrets")
    assert config.path_config_file == "secrets/path_config_file"


# @pytest.mark.current
@pytest.mark.case_sensitive
@pytest.mark.secret
def test_case_sensitive_secrets_dir(secrets_dir):
    path_config_file = secrets_dir / "path_config_file"
    path_config_file.touch(exist_ok=True)
    path_config_file.write_text("secrets/path_config_file")

    PATH_CONFIG_FILE = secrets_dir / "PATH_CONFIG_FILE"
    PATH_CONFIG_FILE.touch(exist_ok=True)
    PATH_CONFIG_FILE.write_text("secrets/PATH_CONFIG_FILE")

    Path_Config_File = secrets_dir / "Path_Config_File"
    Path_Config_File.touch(exist_ok=True)
    Path_Config_File.write_text("secrets/Path_Config_File")

    class AppConfig(ArFiSettings):
        path_config_file: str
        PATH_CONFIG_FILE: str
        Path_Config_File: str
        ordered_settings = [
            "secrets",
        ]
        model_config = SettingsConfigDict(
            secrets_dir="secrets",
            ignore_missing=False,
        )

    config = AppConfig()
    assert config.path_config_file == "secrets/path_config_file"
    assert config.PATH_CONFIG_FILE == "secrets/PATH_CONFIG_FILE"
    assert config.Path_Config_File == "secrets/Path_Config_File"

    class AppConfig(ArFiSettings):
        path_config_FILE: str
        ordered_settings = [
            "secrets",
        ]
        model_config = SettingsConfigDict(
            secrets_dir="secrets",
            ignore_missing=False,
        )

    config = AppConfig()
    assert config.path_config_FILE == "secrets/path_config_file"

    with pytest.raises(pydantic.ValidationError) as excinfo:

        class AppConfig(ArFiSettings):
            path_config_FILE: str
            ordered_settings = [
                "secrets",
            ]
            model_config = SettingsConfigDict(
                secrets_dir="secrets",
                ignore_missing=False,
                case_sensitive=True,
            )

        _ = AppConfig()

    assert excinfo.value.errors(include_url=False) == [
        {
            "type": "missing",
            "loc": ("path_config_FILE",),
            "msg": "Field required",
            "input": {},
        },
    ]


# @pytest.mark.current
@pytest.mark.alias
@pytest.mark.secret
def test_alias_choices_secrets_dir(secrets_dir):
    path_config_file = secrets_dir / "path_config_file"
    path_config_file.touch(exist_ok=True)
    path_config_file.write_text("secrets/path_config_file")

    PATH_CONFIG_FILE = secrets_dir / "PATH_CONFIG_FILE"
    PATH_CONFIG_FILE.touch(exist_ok=True)
    PATH_CONFIG_FILE.write_text("secrets/PATH_CONFIG_FILE")

    Path_Config_File = secrets_dir / "Path_Config_File"
    Path_Config_File.touch(exist_ok=True)
    Path_Config_File.write_text("secrets/Path_Config_File")

    class AppConfig(ArFiSettings):
        path: str = Field(alias="path_config_file")
        ordered_settings = [
            "secrets",
        ]
        model_config = SettingsConfigDict(
            secrets_dir="secrets",
            ignore_missing=False,
        )

    config = AppConfig()
    assert config.path == "secrets/path_config_file"

    class AppConfig(ArFiSettings):
        path: str = Field(
            validation_alias=AliasChoices("path_CONFIG_file", "path_config_FILE"),
        )
        ordered_settings = [
            "secrets",
        ]
        model_config = SettingsConfigDict(
            secrets_dir="secrets",
            ignore_missing=False,
        )

    config = AppConfig()
    assert config.path == "secrets/path_config_file"

    class AppConfig(ArFiSettings):
        path: str = Field(
            validation_alias=AliasChoices("Path_Config_File", "path_config_file"),
        )
        ordered_settings = [
            "secrets",
        ]
        model_config = SettingsConfigDict(
            secrets_dir="secrets",
            ignore_missing=False,
        )

    config = AppConfig()
    assert config.path == "secrets/Path_Config_File"


# @pytest.mark.current
@pytest.mark.alias
@pytest.mark.secret
def test_alias_path_secrets_dir(secrets_dir):
    path_for_alias = secrets_dir / "path_for_alias"
    path_for_alias.touch(exist_ok=True)
    path_for_alias.write_text(
        '["secrets/path_for_alias_0","secrets/path_for_alias_1"]',
    )

    class AppConfig(ArFiSettings):
        path: str = Field(
            validation_alias=AliasPath("path_for_alias", 1),
        )
        ordered_settings = [
            "secrets",
        ]
        model_config = SettingsConfigDict(
            secrets_dir="secrets",
            ignore_missing=False,
        )

    config = AppConfig()
    assert config.path == "secrets/path_for_alias_1"



import functools
import inspect
import warnings
from pathlib import Path

import pydantic
from pydantic import BaseModel, ConfigDict, Field
from pydantic_core import PydanticUndefined
from typing_extensions import Annotated

from .constants import (
    PYPROJECT_TOML_MAX_DEPTH,
)
from .readers import ArFiReader
from .schemes import PyProjectSchema
from .types import PathType
from .utils import search_dict_for_path

__all__ = [
    "init_settings",
]


class InitSettings(BaseModel):
    """Init settings."""

    root_dir: PathType | None = None
    base_dir: PathType | None = None
    pyproject_toml_path: PathType | None = None
    pyproject_toml_depth: Annotated[int, Field(ge=0)] | None = None
    pyproject_toml_max_depth: Annotated[int, Field(ge=0)] | None = PYPROJECT_TOML_MAX_DEPTH
    init_params: PyProjectSchema = PyProjectSchema()
    main_config_file: Path | None = None
    read_pyproject_toml: bool = True
    called_line: int | None = None
    main_config_class: str | None = None

    model_config = ConfigDict(
        extra="forbid",
        validate_assignment=True,
        validate_default=False,
        validate_return=False,
    )

    def read_pyproject(
        self,
        read_once: bool = False,
        read_pyproject_toml: bool = True,
        pyproject_toml_depth: int | None = None,
        pyproject_toml_max_depth: int = PYPROJECT_TOML_MAX_DEPTH,
        class_name: str | None = None,
        search_base_dir: bool = True,
    ):
        """Searches and read pyproject.toml file."""

        if not read_pyproject_toml:
            return
        self.pyproject_toml_depth = pyproject_toml_depth
        self.pyproject_toml_max_depth = pyproject_toml_max_depth

        called_file, lineno = self._search_called_file()
        self.main_config_class = class_name
        self.called_line = lineno
        if self.main_config_file is not None and called_file == self.main_config_file.as_posix():
            return

        self.main_config_file = Path(called_file).resolve()
        if not self.read_pyproject_toml:
            return
        if read_once:
            self.read_pyproject_toml = False
        pyproject_toml_path = self._search_pyproject_toml(
            called_file,
            pyproject_toml_depth,
            pyproject_toml_max_depth,
        )
        if pyproject_toml_path is None:
            return
        if pyproject_toml_path == self.pyproject_toml_path:
            return

        if self.pyproject_toml_path is not None:
            warnings.warn_explicit(
                f"\033[33m\n"
                f"Path to pyproject.toml has been changed !!!\n"
                f"instance {class_name}()\n"
                f"    previous path:\n"
                f"{self.pyproject_toml_path.as_posix()}\n"
                f"    current path:\n"
                f"{pyproject_toml_path.as_posix()}\n"
                f"Call once\n"
                f"  from arfi_settings.init_config import init_settings\n"
                f"  init_settings.read_pyproject(read_once=True)\n"
                f"before import any instance or subclass `ArFiSettings` for fix it."
                f"\033[0m",
                category=Warning,
                filename=called_file,
                lineno=lineno,
            )

        self.pyproject_toml_path = pyproject_toml_path
        self.root_dir = pyproject_toml_path.parent.resolve()

        reader = ArFiReader(
            file_path=pyproject_toml_path,
        )
        pyproject_toml_data = reader.read()
        search_path = ["tool", "arfi_settings"]
        init_data = search_dict_for_path(search_path, pyproject_toml_data)
        if init_data is not PydanticUndefined:
            if init_data and isinstance(init_data, dict):
                try:
                    self.init_params = PyProjectSchema(**init_data)
                except pydantic.ValidationError as e:
                    e.add_note(f"Source file: \033[32m{pyproject_toml_path.as_posix()}\033[0m")
                    raise e

    @staticmethod
    def _search_called_file() -> tuple[str, int]:
        """Searches file which called initialization ArFiSettings."""

        called_file = ""
        lineno = 0
        for elem in reversed(inspect.stack()):
            if "importlib" in elem.filename:
                continue
            if elem.filename.endswith("/arfi_settings/main.py"):
                break
            if "/arfi_settings/" in elem.filename:
                continue
            lineno = elem.lineno
            called_file = Path(elem.filename).resolve().as_posix()

        return called_file, lineno

    def _search_pyproject_toml(
        self,
        cwd: PathType,
        depth: int | None = None,
        max_depth: int | None = PYPROJECT_TOML_MAX_DEPTH,
    ) -> Path | None:
        """Search pyproject.toml file."""

        pyproject_toml_file = None
        if isinstance(cwd, str):
            cwd = Path(cwd)
        assert isinstance(cwd, Path)

        if cwd.is_file():
            cwd = cwd.parent

        assert cwd.is_dir()

        if max_depth is None:
            max_depth = PYPROJECT_TOML_MAX_DEPTH

        real_depth = None
        if depth is not None:
            real_depth = depth
            self.pyproject_toml_max_depth = None
            if depth:
                for _ in range(0, depth):
                    cwd = cwd.parent
        elif max_depth:
            for i in range(0, max_depth + 1):
                file_path = cwd / "pyproject.toml"
                if file_path.is_file():
                    real_depth = i
                    break
                else:
                    if i < max_depth:
                        cwd = cwd.parent

        file_path = cwd / "pyproject.toml"
        if file_path.is_file():
            pyproject_toml_file = file_path.resolve()
            self.pyproject_toml_depth = real_depth
        return pyproject_toml_file

    def _search_base_dir(self):
        """Searches base dir for project."""

        current_root_dir = self.main_config_file.parent
        while self.root_dir is None:
            if Path(current_root_dir / "__init__.py").is_file():
                current_root_dir = current_root_dir.parent
                continue
            self.root_dir = current_root_dir.resolve()
        base_dir = self.main_config_file.parent
        previous_dir = base_dir
        while previous_dir != self.root_dir:
            if not Path(previous_dir / "__init__.py").is_file():
                break
            base_dir = previous_dir
            previous_dir = previous_dir.parent

        if Path(previous_dir / "__init__.py").is_file():
            base_dir = previous_dir

        self.base_dir = base_dir.resolve()

    def _search_base_dir_decorator(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            func(*args, **kwargs)
            if kwargs.get("search_base_dir"):
                self._search_base_dir()

        return wrapper

    def __getattribute__(self, name):
        value = object.__getattribute__(self, name)
        if name == "read_pyproject":
            value = self._search_base_dir_decorator(value)
        return value


init_settings: InitSettings = InitSettings()



import pytest
from pydantic import Field, AliasChoices, AliasPath, ValidationError

from arfi_settings import ArFiSettings, SettingsConfigDict


# @pytest.mark.current
@pytest.mark.alias
@pytest.mark.parametrize(
    "data",
    [
        {"app": {"mode": "asd"}},
        {"app": {"MODE": "asd"}},
    ],
)
def test_mode_alias(data):
    class AppSettings(ArFiSettings):
        pass

    class AppConfig(ArFiSettings):
        app: AppSettings

    config = AppConfig(**data)
    assert config.MODE is None
    assert config.app.MODE == "asd"


# @pytest.mark.current
@pytest.mark.alias
def test_without_aliase(alias_data_config_config_toml, path_base_dir):
    class AppConfig(ArFiSettings):
        read_config_force = False
        alias_config_data: str = "default"

    config = AppConfig()
    assert config.alias_config_data == "default"

    class AppConfig(ArFiSettings):
        alias_config_data: str = "default"

    config = AppConfig()
    assert config.alias_config_data == "alias_config_data"


# @pytest.mark.current
@pytest.mark.alias
def test_strict_aliase(alias_data_config_config_toml, path_base_dir):
    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(
            "default",
            alias="ALIAS_CONFIG_DATA",
        )

    config = AppConfig()
    assert config.alias_config_data == "ALIAS_CONFIG_DATA"


# @pytest.mark.current
@pytest.mark.alias
def test_validation_aliase(alias_data_config_config_toml, path_base_dir):
    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(
            "default",
            alias="ALIAS_CONFIG_DATA",
            validation_alias="AliasConfigData",
        )

    config = AppConfig()
    assert config.alias_config_data == "AliasConfigData"


@pytest.mark.alias
def test_validation_aliase_aliaschoices(alias_data_config_config_toml, path_base_dir):
    class AppConfig(ArFiSettings):
        # read_config_force = False
        alias_config_data: str = Field(
            "default",
            alias="Alias_Config_Data",
            validation_alias=AliasChoices("configData", "AliasConfigData"),
        )

    config = AppConfig()
    assert config.alias_config_data == "configData"


@pytest.mark.alias
def test_validation_aliase_multi_aliaspath(alias_data_config_config_toml, path_base_dir):
    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(
            "default",
            validation_alias=AliasPath("complex_alias_path", "alias_path", 0),
        )

    config = AppConfig()
    assert config.alias_config_data == "alias_path_1"


@pytest.mark.alias
def test_validation_aliase_aliaspath(alias_data_config_config_toml, path_base_dir):
    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(
            "default",
            alias="Alias_Config_Data",
            validation_alias=AliasPath("alias_path", 0),
        )

    config = AppConfig()
    assert config
    assert config.alias_config_data == "alias_path_1"


# @pytest.mark.current
@pytest.mark.alias
def test_validation_aliase_with_extra(alias_data_config_config_toml, path_base_dir):
    data = {"alias_path": ["alias_path_1", "alias_path_2"], "asd": "alias_path_2"}

    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(validation_alias=AliasPath("alias_path", 0))
        asd: str = Field(validation_alias=AliasChoices("asd", AliasPath("alias_path", 1)))

    config = AppConfig(**data)
    assert config
    assert config.alias_config_data == "alias_path_1"
    assert config.asd == "alias_path_2"

    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(validation_alias=AliasPath("alias_path", 0))
        asd: str = Field(validation_alias=AliasChoices(AliasPath("alias_path", 1), "asd"))

    config = AppConfig(**data)
    assert config
    assert config.alias_config_data == "alias_path_1"
    assert config.asd == "alias_path_2"

    class AppConfig(ArFiSettings):
        read_config_force = False
        alias_config_data: str = Field(validation_alias=AliasPath("alias_path", 0))
        asd: str = Field(validation_alias=AliasChoices("asd", AliasPath("alias_path", 1)))

    config = AppConfig(**data)
    assert config
    assert config.alias_config_data == "alias_path_1"
    assert config.asd == "alias_path_2"

    with pytest.raises(ValidationError) as excinfo:

        class AppConfig(ArFiSettings):
            read_config_force = False
            alias_config_data: str = Field(validation_alias=AliasPath("alias_path", 0))
            asd: str = Field(validation_alias=AliasChoices(AliasPath("alias_path", 1), "asd"))

        _ = AppConfig(**data)
    assert excinfo.value.errors(include_url=False) == [
        {
            "type": "extra_forbidden",
            "loc": ("asd",),
            "msg": "Extra inputs are not permitted",
            "input": "alias_path_2",
        },
    ]

    class AppConfig(ArFiSettings):
        read_config_force = False
        alias_config_data: str = Field(validation_alias=AliasPath("alias_path", 0))
        asd: str = Field(validation_alias=AliasChoices(AliasPath("alias_path", 1), "asd"))

        model_config = SettingsConfigDict(
            extra="ignore",
        )

    config = AppConfig(**data)
    assert config
    assert config.alias_config_data == "alias_path_1"
    assert config.asd == "alias_path_2"


# @pytest.mark.current
@pytest.mark.alias
@pytest.mark.parametrize(
    "validation_alias_config, validation_alias_asd, data, res_alias_config_data, res_asd",
    [
        (
            AliasPath("alias_path", 0),
            AliasChoices("asd", AliasPath("alias_path", 1)),
            {},
            "alias_path_1",
            "alias_path_2",
        ),
        (
            AliasPath("alias_path"),
            AliasChoices("asd", AliasPath("alias_path", 1)),
            {},
            ["alias_path_1", "alias_path_2"],
            "alias_path_2",
        ),
        (
            AliasPath("complex_alias_path", "alias_path"),
            AliasChoices("asd", AliasPath("alias_path", 1)),
            {},
            ["alias_path_1", "alias_path_2"],
            "alias_path_2",
        ),
        (
            AliasPath("alias_path", 0),
            AliasChoices("asd", AliasPath("alias_path", 1)),
            {"alias_path": ["alias_path_1", "alias_path_2"]},
            "alias_path_1",
            "alias_path_2",
        ),
        (
            AliasPath("alias_path", 0),
            AliasChoices("asd", AliasPath("alias_path", 1)),
            {"alias_path": ["alias_path_1"], "asd": "alias_path_8"},
            "alias_path_1",
            "alias_path_8",
        ),
        (
            AliasPath("alias_path", 0),
            AliasChoices(AliasPath("alias_path", 1), "asd"),
            {"alias_path": ["alias_path_1"], "asd": "alias_path_8"},
            "alias_path_1",
            "alias_path_8",
        ),
        (
            AliasPath("alias_path", 0),
            AliasChoices(AliasPath("alias_path", 1), "asd"),
            {"alias_path": ["alias_path_1"]},
            "alias_path_1",
            "alias_path_2",
        ),
        (
            AliasPath("alias_path", 0),
            AliasChoices("asd", AliasPath("alias_path", 1)),
            {"alias_path": ["alias_path_1", "alias_path_2"], "asd": "alias_path_8"},
            "alias_path_1",
            "alias_path_8",
        ),
        (
            AliasPath("alias_path", 0),
            AliasChoices(AliasPath("alias_path", 1), "asd"),
            {"alias_path": ["alias_path_1", "alias_path_2"], "asd": "alias_path_8"},
            "alias_path_1",
            "alias_path_2",
        ),
        (
            AliasChoices("alias_config_data", "Alias_Config_Data"),
            AliasChoices("asd", AliasPath("alias_path", 1)),
            {},
            "alias_config_data",
            "alias_path_2",
        ),
        (
            AliasChoices("alias_config_data", "Alias_Config_Data"),
            AliasChoices("asd", AliasPath("alias_path", 1)),
            {"alias_config_data": "Alias_Config_Data"},
            "Alias_Config_Data",
            "alias_path_2",
        ),
        (
            AliasChoices("Alias_Config_Data", "alias_config_data"),
            AliasChoices("asd", AliasPath("alias_path", 1)),
            {"alias_config_data": "Alias_Config_Data"},
            "Alias_Config_Data",
            "alias_path_2",
        ),
        (
            AliasChoices("alias_config_data", "Alias_Config_Data"),
            AliasChoices("asd", AliasPath("alias_path", 1)),
            {"Alias_Config_Data": "Alias_Config_Data"},
            "Alias_Config_Data",
            "alias_path_2",
        ),
    ],
)
def test_validation_aliase_aliaschoices_aliaspath(
    alias_data_config_config_toml,
    validation_alias_config,
    validation_alias_asd,
    data,
    res_alias_config_data,
    res_asd,
    path_base_dir,
):
    class AppConfig(ArFiSettings):
        alias_config_data: str | list | dict = Field(validation_alias=validation_alias_config)
        asd: str = Field(validation_alias=validation_alias_asd)

    config = AppConfig(**data)
    assert config.alias_config_data == res_alias_config_data
    assert config.asd == res_asd


# @pytest.mark.current
@pytest.mark.alias
def test_validation_aliase_aliaschoices_aliaspath_on_default_valut(
    alias_data_config_config_toml,
    path_base_dir,
):
    class AppSettings(ArFiSettings):
        alias_config_data: str = Field(validation_alias=AliasPath("alias_path", 0))
        asd: str = Field(validation_alias=AliasChoices("asd", AliasPath("alias_path", 1)))

    class AppConfig(ArFiSettings):
        app: AppSettings

    app = AppSettings()
    config = AppConfig(app=app)
    assert config.app.alias_config_data == "alias_path_1"
    assert config.app.asd == "alias_path_2"

    config = AppConfig(app=AppSettings())
    assert config.app.alias_config_data == "alias_path_1"
    assert config.app.asd == "alias_path_2"

    config = AppConfig(app=AppSettings(alias_path=["alias_path_1"]))
    assert config.app.alias_config_data == "alias_path_1"
    assert config.app.asd == "alias_path_2"

    config = AppConfig(app=AppSettings(alias_path=["alias_path_1"], asd="alias_path_8"))
    assert config.app.alias_config_data == "alias_path_1"
    assert config.app.asd == "alias_path_8"

    config = AppConfig(app=AppSettings(alias_path=["alias_path_1", "alias_path_2"], asd="alias_path_8"))
    assert config.app.alias_config_data == "alias_path_1"
    assert config.app.asd == "alias_path_8"

    config = AppConfig(app=AppSettings(alias_config_data="alias_path_1", asd="alias_path_8"))
    assert config.app.alias_config_data == "alias_path_1"
    assert config.app.asd == "alias_path_8"

    with pytest.raises(ValidationError) as excinfo:

        class AppSettings(ArFiSettings):
            read_config_force = False
            alias_config_data: str = Field("default1", validation_alias=AliasPath("alias_path", 0))
            asd: str = Field("default2", validation_alias=AliasChoices("asd", AliasPath("alias_path", 1)))

        _ = AppConfig(app=AppSettings(alias_config_data="alias_path_1", asd="alias_path_8"))
    assert excinfo.value.errors(include_url=False) == [
        {
            "type": "extra_forbidden",
            "loc": ("alias_config_data",),
            "msg": "Extra inputs are not permitted",
            "input": "alias_path_1",
        },
    ]

    class AppSettings(ArFiSettings):
        read_config_force = False
        alias_config_data: str = Field("default1", validation_alias=AliasPath("alias_path", 0))
        asd: str = Field("default2", validation_alias=AliasChoices("asd", AliasPath("alias_path", 1)))
        model_config = SettingsConfigDict(extra="ignore")

    config = AppConfig(app=AppSettings(alias_config_data="alias_path_1", asd="alias_path_8"))
    assert config.app.alias_config_data == "default1"
    assert config.app.asd == "alias_path_8"

    config = AppConfig(app=AppSettings(alias_path=["alias_path_9"], asd="alias_path_8"))
    assert config.app.alias_config_data == "alias_path_9"
    assert config.app.asd == "alias_path_8"

    config = AppConfig(
        app=AppSettings(
            alias_config_data="alias_path_1",
            alias_path=["alias_path_9"],
            asd="alias_path_8",
        )
    )
    assert config.app.alias_config_data == "alias_path_9"
    assert config.app.asd == "alias_path_8"

    app = AppSettings(alias_path=["alias_path_1"], asd="alias_path_8")
    config = AppConfig(app=app)
    assert config.app.alias_config_data == "alias_path_1"
    assert config.app.asd == "alias_path_8"

    test_app = AppSettings(alias_path=["alias_path_1"], asd="alias_path_8")

    class AppConfig(ArFiSettings):
        app: AppSettings = test_app

    config = AppConfig()
    assert config.app.alias_config_data == "alias_path_1"
    assert config.app.asd == "alias_path_8"

    config = AppConfig(app=AppSettings(alias_path=["alias_path_1"], asd="ALIAS_PATH_5"))
    assert config.app.alias_config_data == "alias_path_1"
    assert config.app.asd == "ALIAS_PATH_5"


# @pytest.mark.current
@pytest.mark.alias
def test_alias_generator(alias_data_config_config_toml, path_base_dir):
    def to_camel(string: str):
        return "".join(x.capitalize() for x in string.split("_"))

    class AppConfig(ArFiSettings):
        alias_config_data: str
        model_config = SettingsConfigDict(alias_generator=to_camel)

    config = AppConfig()
    assert config.alias_config_data == "AliasConfigData"

    config = AppConfig(alias_config_data="alias_config_data")
    assert config.alias_config_data == "AliasConfigData"

    config = AppConfig(AliasConfigData="alias_config_data")
    assert config.alias_config_data == "alias_config_data"

    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(
            validation_alias=AliasChoices(
                "Alias_Config_Data",
                "alias_config_data",
            ),
        )
        model_config = SettingsConfigDict(alias_generator=to_camel)

    config = AppConfig()
    assert config.alias_config_data == "Alias_Config_Data"

    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(
            alias="Alias_Config_Data",
            validation_alias=AliasChoices(
                "CONFIG_DATA",
            ),
        )
        model_config = SettingsConfigDict(alias_generator=to_camel)

    config = AppConfig()
    assert config.alias_config_data == "CONFIG_DATA"

    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(
            alias_priority=1,
            alias="Alias_Config_Data",
            validation_alias=AliasChoices(
                "configData",
            ),
        )
        model_config = SettingsConfigDict(alias_generator=to_camel)

    config = AppConfig()
    assert config.alias_config_data == "AliasConfigData"

    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(
            alias_priority=2,
            alias="Alias_Config_Data",
            validation_alias=AliasChoices(
                "configData",
            ),
        )
        model_config = SettingsConfigDict(alias_generator=to_camel)

    config = AppConfig()
    assert config.alias_config_data == "configData"

    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(
            alias_priority=30,
            alias="Alias_Config_Data",
            validation_alias=AliasChoices(
                "configData",
            ),
        )
        model_config = SettingsConfigDict(alias_generator=to_camel)

    config = AppConfig()
    assert config.alias_config_data == "configData"

    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(
            alias_priority=1,
            alias="configData",
        )
        model_config = SettingsConfigDict(alias_generator=to_camel)

    config = AppConfig()
    assert config.alias_config_data == "AliasConfigData"

    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(
            alias_priority=2,
            alias="configData",
        )
        model_config = SettingsConfigDict(alias_generator=to_camel)

    config = AppConfig()
    assert config.alias_config_data == "configData"

    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(
            alias_priority=30,
            alias="configData",
        )
        model_config = SettingsConfigDict(alias_generator=to_camel)

    config = AppConfig()
    assert config.alias_config_data == "configData"

    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(alias_priority=0)
        model_config = SettingsConfigDict(alias_generator=to_camel)

    config = AppConfig()
    assert config.alias_config_data == "AliasConfigData"

    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(alias_priority=1)
        model_config = SettingsConfigDict(alias_generator=to_camel)

    config = AppConfig()
    assert config.alias_config_data == "AliasConfigData"

    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(alias_priority=2)
        model_config = SettingsConfigDict(alias_generator=to_camel)

    config = AppConfig()
    assert config.alias_config_data == "AliasConfigData"

    class AppConfig(ArFiSettings):
        alias_config_data: str = Field(alias_priority=30)
        model_config = SettingsConfigDict(alias_generator=to_camel)

    config = AppConfig()
    assert config.alias_config_data == "AliasConfigData"



import pydantic
import pytest
from pydantic import StrictBool, validate_call

from arfi_settings import ArFiSettings
from arfi_settings.descriptors import (
    ABCDescriptor,
    EnvConfigInheritParentDescriptor,
    FileConfigInheritParentDescriptor,
    ModeDirDescriptor,
    ModeDirInheritNestedDescriptor,
    ModeDirInheritParentDescriptor,
    ReadConfigDeskriptor,
    ReadConfigForceDeskriptor,
)
from arfi_settings.errors import ArFiSettingsError
from arfi_settings.types import DEFAULT_PATH_SENTINEL, SENTINEL


@pytest.mark.descriptors
def test_new_valid_descriptor():
    class NewDescriptor(ABCDescriptor):
        name: str = "attr_name"
        default: bool = True

        @staticmethod
        @validate_call
        def validate(value: StrictBool):
            return value

    new_descriptor = NewDescriptor()
    assert new_descriptor.name == "attr_name"
    assert new_descriptor.default is True
    with pytest.raises(ArFiSettingsError) as excinfo:
        _ = new_descriptor.inherited_value
    assert "Can not define `inherited_value` without owner" in str(excinfo.value)


@pytest.mark.descriptors
def test_new_descriptors_without_validate_method():
    with pytest.raises(TypeError):

        class NewDescriptor(ABCDescriptor):
            name: str = "attr_name"
            default: bool = True

        new_descriptor = NewDescriptor()
        assert new_descriptor.name == "test_name"


@pytest.mark.descriptors
@pytest.mark.parametrize(
    "attr_name, default_value, error_message",
    [
        (SENTINEL, True, "Missing or empty `name` attribut"),
        ("test_name", SENTINEL, "Missing `default` attribut"),
        (None, True, "attribut `name` must be not empty string"),
        ("", True, "attribut `name` must be not empty string"),
    ],
)
def test_new_descriptor_name_and_default_value(attr_name, default_value, error_message):
    with pytest.raises(ArFiSettingsError) as excinfo:

        class NewDescriptor(ABCDescriptor):
            name: str = attr_name
            default: bool = default_value

            @staticmethod
            @validate_call
            def validate(value: StrictBool):
                return value

    assert error_message in str(excinfo.value)


@pytest.mark.descriptors
def test_validate_init_default_value_descriptor():
    class NewDescriptor(ABCDescriptor):
        name: str = "attr_name"
        default: bool = True

        @staticmethod
        @validate_call
        def validate(value: StrictBool):
            return value

    new_descriptor = NewDescriptor(True)
    assert new_descriptor.default is True
    with pytest.raises(pydantic.ValidationError):
        new_descriptor = NewDescriptor(1)
        assert new_descriptor.default is True


@pytest.mark.descriptors
def test_read_config_descriptor():
    read_config = ReadConfigDeskriptor()
    assert read_config.name == "read_config"
    assert read_config.default is True

    class AppConfig:
        read_config = ReadConfigDeskriptor(False)

    config = AppConfig()
    assert not config.read_config


@pytest.mark.descriptors
def test_read_config_force_descriptor():
    read_config_force = ReadConfigForceDeskriptor()
    assert read_config_force.name == "read_config_force"
    assert read_config_force.default is None

    class AppConfig:
        read_config_force = ReadConfigDeskriptor(True)

    config = AppConfig()
    assert config.read_config_force is True

    class AppConfig:
        read_config_force = ReadConfigDeskriptor(False)

    config = AppConfig()
    assert config.read_config_force is False


@pytest.mark.descriptors
def test_mode_dir_inherit_nested_descriptor():
    mode_dir_inherit_nested = ModeDirInheritNestedDescriptor()
    assert mode_dir_inherit_nested.name == "mode_dir_inherit_nested"
    assert mode_dir_inherit_nested.default is True

    class AppConfig:
        _mode_dir_inherit_nested = True
        mode_dir_inherit_nested = ModeDirInheritNestedDescriptor(False)

    config = AppConfig()
    assert config.mode_dir_inherit_nested


@pytest.mark.descriptors
def test_mode_dir_inherit_parent_descriptor():
    mode_dir_inherit_parent = ModeDirInheritParentDescriptor()
    assert mode_dir_inherit_parent.name == "mode_dir_inherit_parent"
    assert mode_dir_inherit_parent.default is True

    class AppConfig:
        _mode_dir_inherit_parent = True
        mode_dir_inherit_parent = ModeDirInheritParentDescriptor(False)

    config = AppConfig()
    assert config.mode_dir_inherit_parent


@pytest.mark.descriptors
def test_file_config_inherit_parent_descriptor():
    file_config_inherit_parent = FileConfigInheritParentDescriptor()
    assert file_config_inherit_parent.name == "file_config_inherit_parent"
    assert file_config_inherit_parent.default is True

    class AppConfig:
        _file_config_inherit_parent = True
        file_config_inherit_parent = FileConfigInheritParentDescriptor(False)

    config = AppConfig()
    assert config.file_config_inherit_parent


@pytest.mark.descriptors
def test_env_config_inherit_parent_descriptor():
    env_config_inherit_parent = EnvConfigInheritParentDescriptor()
    assert env_config_inherit_parent.name == "env_config_inherit_parent"
    assert env_config_inherit_parent.default is True

    class AppConfig:
        _env_config_inherit_parent = True
        env_config_inherit_parent = EnvConfigInheritParentDescriptor(False)

    config = AppConfig()
    assert config.env_config_inherit_parent


@pytest.mark.descriptors
def test_mode_dir_descriptor():
    mode_dir = ModeDirDescriptor()
    assert mode_dir.name == "mode_dir"
    assert mode_dir.default is DEFAULT_PATH_SENTINEL

    class AppConfig:
        mode_dir_inherit_nested = True
        mode_dir = ModeDirDescriptor()

    config = AppConfig()
    assert config.mode_dir == DEFAULT_PATH_SENTINEL

    class AppConfig:
        mode_dir_inherit_nested = True
        _mode_dir = "test"
        mode_dir = ModeDirDescriptor()

    config = AppConfig()
    assert config.mode_dir == "test"


@pytest.mark.descriptors
@pytest.mark.parametrize(
    "self_mode_dir, parent_class_attr, result_mode_dir",
    [
        ("test", {}, "test"),
        ("test", {"mode_dir": "parent"}, "parent/test"),
        (DEFAULT_PATH_SENTINEL, {"mode_dir": "parent"}, "parent"),
        (None, {"mode_dir": "parent"}, "parent"),
        ("test", {"mode_dir": DEFAULT_PATH_SENTINEL}, "test"),
        ("test", {"mode_dir": SENTINEL}, "test"),
    ],
)
def test_parent_mode_dir_descriptor(self_mode_dir, parent_class_attr, result_mode_dir):
    Parent = type("Parent", (), parent_class_attr)

    class AppConfig(Parent):
        mode_dir_inherit_nested = True
        mode_dir = ModeDirDescriptor(self_mode_dir)

    config = AppConfig()
    assert config.mode_dir == result_mode_dir


@pytest.mark.descriptors
def test_mode_dir_missing_attr_descriptor():
    with pytest.raises(ArFiSettingsError) as excinfo:

        class AppConfig:
            mode_dir = ModeDirDescriptor()

        config = AppConfig()
        assert config.mode_dir == DEFAULT_PATH_SENTINEL
    assert "missing required attribute `mode_dir_inherit_nested`" in str(excinfo.value)


@pytest.mark.descriptors
@pytest.mark.parametrize(
    "wrong_mode_dir, error_message",
    [
        (".", "`mode_dir` can't startswith `.`"),
        ("/config", "`mode_dir` can't startswith `/`"),
        ("config/", "mode_dir` can't endswith `/`"),
    ],
)
def test_wrong_mode_dir_descriptor(wrong_mode_dir, error_message):
    with pytest.raises(ArFiSettingsError) as excinfo:

        class AppConfig:
            mode_dir_inherit_nested = True
            mode_dir = ModeDirDescriptor(wrong_mode_dir)

        config = AppConfig()
        assert config.mode_dir == wrong_mode_dir

    assert error_message in str(excinfo.value)


@pytest.mark.descriptors
def test_abc_descriptor():
    from typing import Any

    class NewDescriptor(ABCDescriptor):
        name = "test"
        default = 1

        @staticmethod
        def validate(value: Any):
            return value

    new_decriptor = NewDescriptor()
    assert new_decriptor.__get__("abstract_instance") == new_decriptor
    assert new_decriptor.validate("Any") == "Any"


@pytest.mark.descriptors
@pytest.mark.parametrize(
    """
    _read_config,
    _mode_dir_inherit_nested,
    _mode_dir_inherit_parent,
    _file_config_inherit_parent,
    _env_config_inherit_parent
    """,
    [
        ("True", True, True, True, True),
        (True, "True", True, True, True),
        (True, True, "True", True, True),
        (True, True, True, "True", True),
        (True, True, True, True, "True"),
    ],
)
def test_not_bool_value_descriptor(
    _read_config,
    _mode_dir_inherit_nested,
    _mode_dir_inherit_parent,
    _file_config_inherit_parent,
    _env_config_inherit_parent,
):
    class Proxy(ArFiSettings):
        pass

    class AppSettings(ArFiSettings):
        proxy: Proxy = Proxy()

    with pytest.raises(pydantic.ValidationError):

        class AppConfig(ArFiSettings):
            read_config = _read_config
            mode_dir_inherit_nested = _mode_dir_inherit_nested
            mode_dir_inherit_parent = _mode_dir_inherit_parent
            file_config_inherit_parent = _file_config_inherit_parent
            env_config_inherit_parent = _env_config_inherit_parent
            app: AppSettings = AppSettings()


###
